@@ -25,7 +25,11 @@ fi
 }
 
 echo Running autoreconf --verbose --install --warnings=all
-autoreconf --verbose --install --warnings=all
+autoreconf --force --verbose --install --warnings=all
+
+# This gets around the BitKeeper problem that it checks out files with the
+# current time stamp rather than the time stamp at check in time.
+touch config.h.in
 
 if test -z "$*"; then
   echo "**Warning**: I am going to run \`configure' with no arguments."
@@ -34,11 +38,8 @@ if test -z "$*"; then
   echo
 fi
 
-conf_flags="--enable-maintainer-mode --enable-compile-warnings"
-
-echo Running $srcdir/configure $conf_flags "$@" ...
-$srcdir/configure $conf_flags "$@" \
-	&& echo Now type \`make\' to compile ntfsprogs. || err=1
+echo Running $srcdir/configure "$@" ...
+$srcdir/configure "$@" && echo Now type \`make\' to compile ntfsprogs. || err=1
 
 if test "$srcdir" != "."; then
 	popd > /dev/null
