@@ -21,9 +21,15 @@
 
 #include "config.h"
 
-#include <stdlib.h>
-#include <errno.h>
-#include <string.h>
+#ifdef HAVE_STDLIB_H
+#	include <stdlib.h>
+#endif
+#ifdef HAVE_ERRNO_H
+#	include <errno.h>
+#endif
+#ifdef HAVE_STRING_H
+#	include <string.h>
+#endif
 
 #include "types.h"
 #include "debug.h"
@@ -35,7 +41,7 @@
 /*
  * The little endian Unicode string "$I30" as a global constant.
  */
-uchar_t I30[5] = { const_cpu_to_le16('$'), const_cpu_to_le16('I'),
+ntfschar I30[5] = { const_cpu_to_le16('$'), const_cpu_to_le16('I'),
 		   const_cpu_to_le16('3'), const_cpu_to_le16('0'),
 		   const_cpu_to_le16('\0') };
 
@@ -64,7 +70,7 @@ uchar_t I30[5] = { const_cpu_to_le16('$'), const_cpu_to_le16('I'),
  * If the volume is mounted with the case sensitive flag set, then we only
  * allow exact matches.
  */
-u64 ntfs_inode_lookup_by_name(ntfs_inode *dir_ni, const uchar_t *uname,
+u64 ntfs_inode_lookup_by_name(ntfs_inode *dir_ni, const ntfschar *uname,
 		const int uname_len)
 {
 	VCN vcn;
@@ -137,7 +143,7 @@ u64 ntfs_inode_lookup_by_name(ntfs_inode *dir_ni, const uchar_t *uname,
 		 * returning.
 		 */
 		if (ntfs_names_are_equal(uname, uname_len,
-				(uchar_t*)&ie->key.file_name.file_name,
+				(ntfschar*)&ie->key.file_name.file_name,
 				ie->key.file_name.file_name_length,
 				CASE_SENSITIVE, vol->upcase, vol->upcase_len)) {
 found_it:
@@ -158,7 +164,7 @@ found_it:
 		if (!NVolCaseSensitive(vol) &&
 				ie->key.file_name.file_name_type &&
 				ntfs_names_are_equal(uname, uname_len,
-				(uchar_t*)&ie->key.file_name.file_name,
+				(ntfschar*)&ie->key.file_name.file_name,
 				ie->key.file_name.file_name_length,
 				IGNORE_CASE, vol->upcase, vol->upcase_len)) {
 			/* Only one case insensitive matching name allowed. */
@@ -178,7 +184,7 @@ found_it:
 		 * know which way in the B+tree we have to go.
 		 */
 		rc = ntfs_names_collate(uname, uname_len,
-				(uchar_t*)&ie->key.file_name.file_name,
+				(ntfschar*)&ie->key.file_name.file_name,
 				ie->key.file_name.file_name_length, 1,
 				IGNORE_CASE, vol->upcase, vol->upcase_len);
 		/*
@@ -197,7 +203,7 @@ found_it:
 		 * collation.
 		 */
 		rc = ntfs_names_collate(uname, uname_len,
-				(uchar_t*)&ie->key.file_name.file_name,
+				(ntfschar*)&ie->key.file_name.file_name,
 				ie->key.file_name.file_name_length, 1,
 				CASE_SENSITIVE, vol->upcase, vol->upcase_len);
 		if (rc == -1)
@@ -329,7 +335,7 @@ descend_into_child_node:
 		 * returning.
 		 */
 		if (ntfs_names_are_equal(uname, uname_len,
-				(uchar_t*)&ie->key.file_name.file_name,
+				(ntfschar*)&ie->key.file_name.file_name,
 				ie->key.file_name.file_name_length,
 				CASE_SENSITIVE, vol->upcase, vol->upcase_len)) {
 found_it2:
@@ -352,7 +358,7 @@ found_it2:
 		if (!NVolCaseSensitive(vol) &&
 				ie->key.file_name.file_name_type &&
 				ntfs_names_are_equal(uname, uname_len,
-				(uchar_t*)&ie->key.file_name.file_name,
+				(ntfschar*)&ie->key.file_name.file_name,
 				ie->key.file_name.file_name_length,
 				IGNORE_CASE, vol->upcase, vol->upcase_len)) {
 			/* Only one case insensitive matching name allowed. */
@@ -372,7 +378,7 @@ found_it2:
 		 * know which way in the B+tree we have to go.
 		 */
 		rc = ntfs_names_collate(uname, uname_len,
-				(uchar_t*)&ie->key.file_name.file_name,
+				(ntfschar*)&ie->key.file_name.file_name,
 				ie->key.file_name.file_name_length, 1,
 				IGNORE_CASE, vol->upcase, vol->upcase_len);
 		/*
@@ -391,7 +397,7 @@ found_it2:
 		 * collation.
 		 */
 		rc = ntfs_names_collate(uname, uname_len,
-				(uchar_t*)&ie->key.file_name.file_name,
+				(ntfschar*)&ie->key.file_name.file_name,
 				ie->key.file_name.file_name_length, 1,
 				CASE_SENSITIVE, vol->upcase, vol->upcase_len);
 		if (rc == -1)
@@ -456,7 +462,7 @@ close_err_out:
 /*
  * The little endian Unicode string ".." for ntfs_readdir().
  */
-static const uchar_t dotdot[3] = { const_cpu_to_le16('.'),
+static const ntfschar dotdot[3] = { const_cpu_to_le16('.'),
 				   const_cpu_to_le16('.'),
 				   const_cpu_to_le16('\0') };
 
@@ -663,7 +669,7 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 
 		parent_mref = ntfs_mft_get_parent_ref(dir_ni);
 		if (parent_mref == ERR_MREF(-1)) {
-			Dprintf("Parent directory not found: %s\n", errno);
+			Dperror("Parent directory not found");
 			goto dir_err_out;
 		}
 
