@@ -3,6 +3,7 @@
  *
  * Copyright (c) 2002-2004 Anton Altaparmakov
  * Copyright (c) 2002 Richard Russon
+ * Copyright (c) 2004 Yura Pakhuchiy
  *
  * This program/include file is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as published
@@ -887,29 +888,36 @@ mpa_err:
 	if (!attr->lowest_vcn) {
 		VCN max_cluster;
 
-		max_cluster = (sle64_to_cpu(attr->allocated_size) +
+		max_cluster = ((sle64_to_cpu(attr->allocated_size) +
 				vol->cluster_size - 1) >>
-				vol->cluster_size_bits;
+				vol->cluster_size_bits) - 1;
 		/*
-		 * If there is a difference between the highest_vcn and the
-		 * highest cluster, the runlist is either corrupt or, more
-		 * likely, there are more extents following this one.
+		 * A highest_vcn of zero means this is a single extent
+		 * attribute so simply terminate the runlist with LCN_ENOENT).
 		 */
-		if (deltaxcn < --max_cluster) {
-			Dprintf("More extents to follow; deltaxcn = 0x%llx, "
-					"max_cluster = 0x%llx\n",
-					(long long)deltaxcn,
-					(long long)max_cluster);
-			rl[rlpos].vcn = vcn;
-			vcn += rl[rlpos].length = max_cluster - deltaxcn;
-			rl[rlpos].lcn = (LCN)LCN_RL_NOT_MAPPED;
-			rlpos++;
-		} else if (deltaxcn > max_cluster) {
-			Dprintf("Corrupt attribute. deltaxcn = 0x%llx, "
-					"max_cluster = 0x%llx",
-					(long long)deltaxcn,
-					(long long)max_cluster);
-			goto mpa_err;
+		if (deltaxcn) {
+			/*
+			 * If there is a difference between the highest_vcn and
+			 * the highest cluster, the runlist is either corrupt
+			 * or, more likely, there are more extents following
+			 * this one.
+			 */
+			if (deltaxcn < max_cluster) {
+				Dprintf("More extents to follow; deltaxcn = 0x%llx, "
+						"max_cluster = 0x%llx\n",
+						(long long)deltaxcn,
+						(long long)max_cluster);
+				rl[rlpos].vcn = vcn;
+				vcn += rl[rlpos].length = max_cluster - deltaxcn;
+				rl[rlpos].lcn = (LCN)LCN_RL_NOT_MAPPED;
+				rlpos++;
+			} else if (deltaxcn > max_cluster) {
+				Dprintf("Corrupt attribute. deltaxcn = 0x%llx, "
+						"max_cluster = 0x%llx",
+						(long long)deltaxcn,
+						(long long)max_cluster);
+				goto mpa_err;
+			}
 		}
 		rl[rlpos].lcn = (LCN)LCN_ENOENT;
 	} else /* Not the base extent. There may be more extents to follow. */
@@ -1032,8 +1040,9 @@ s64 ntfs_rl_pread(const ntfs_volume *vol, const runlist_element *rl,
 	if (!count)
 		return count;
 	/* Seek in @rl to the run containing @pos. */
-	for (ofs = 0; rl->length && (ofs + rl->length <= pos); rl++)
-		ofs += rl->length;
+	for (ofs = 0; rl->length && (ofs + (rl->length <<
+			vol->cluster_size_bits) <= pos); rl++)
+		ofs += (rl->length << vol->cluster_size_bits);
 	/* Offset in the run at which to begin reading. */
 	ofs = pos - ofs;
 	for (total = 0LL; count; rl++, ofs = 0) {
@@ -1049,7 +1058,7 @@ s64 ntfs_rl_pread(const ntfs_volume *vol, const runlist_element *rl,
 			/* Update counters and proceed with next run. */
 			total += to_read;
 			count -= to_read;
-			(u8*)b += to_read;
+			b = (u8*)b + to_read;
 			continue;
 		}
 		/* It is a real lcn, read it from the volume. */
@@ -1062,7 +1071,7 @@ retry:
 		if (bytes_read > 0) {
 			total += bytes_read;
 			count -= bytes_read;
-			(u8*)b += bytes_read;
+			b = (u8*)b + bytes_read;
 			continue;
 		}
 		/* If the syscall was interrupted, try again. */
@@ -1115,8 +1124,9 @@ s64 ntfs_rl_pwrite(const ntfs_volume *vol, const runlist_element *rl,
 	if (!count)
 		return count;
 	/* Seek in @rl to the run containing @pos. */
-	for (ofs = 0; rl->length && (ofs + rl->length <= pos); rl++)
-		ofs += rl->length;
+	for (ofs = 0; rl->length && (ofs + (rl->length <<
+			vol->cluster_size_bits) <= pos); rl++)
+		ofs += (rl->length << vol->cluster_size_bits);
 	/* Offset in the run at which to begin writing. */
 	ofs = pos - ofs;
 	for (total = 0LL; count; rl++, ofs = 0) {
@@ -1157,7 +1167,7 @@ s64 ntfs_rl_pwrite(const ntfs_volume *vol, const runlist_element *rl,
 			 */
 			total += to_write;
 			count -= to_write;
-			(u8*)b += to_write;
+			b = (u8*)b + to_write;
 			continue;
 		}
 		/* It is a real lcn, write it to the volume. */
@@ -1174,7 +1184,7 @@ retry:
 		if (written > 0) {
 			total += written;
 			count -= written;
-			(u8*)b += written;
+			b = (u8*)b + written;
 			continue;
 		}
 		/* If the syscall was interrupted, try again. */
@@ -1233,9 +1243,10 @@ __inline__ int ntfs_get_nr_significant_bytes(const s64 n)
  * array corresponding to the runlist @rl. This for example allows us to
  * allocate a buffer of the right size when building the mapping pairs array.
  *
- * Return the calculated size in bytes on success. If @rl is NULL return 0.
- * On error, return -1 with errno set to the error code. The following error
- * codes are defined:
+ * If @rl is NULL, just return 1 (for the single terminator byte).
+ *
+ * Return the calculated size in bytes on success.  On error, return -1 with
+ * errno set to the error code.  The following error codes are defined:
  *	EINVAL	- Run list contains unmapped elements. Make sure to only pass
  *		  fully mapped runlists to this function.
  *	EIO	- The runlist is corrupt.
@@ -1247,7 +1258,7 @@ int ntfs_get_size_for_mapping_pairs(const ntfs_volume *vol,
 	int i, rls;
 
 	if (!rl)
-		return 0;
+		return 1;
 	/* Always need the termining zero byte. */
 	rls = 1;
 	for (prev_lcn = i = 0; rl[i].length; i++) {
@@ -1342,8 +1353,10 @@ err_out:
  * @dst. @dst_len is the size of @dst in bytes and it should be at least equal
  * to the value obtained by calling ntfs_get_size_for_mapping_pairs().
  *
- * Return 0 on success or when @rl is NULL. On error, return -1 with errno set
- * to the error code. The following error codes are defined:
+ * If @rl is NULL, just write a single terminator byte to @dst.
+ *
+ * Return 0 on success.  On error, return -1 with errno set to the error code.
+ * The following error codes are defined:
  *	EINVAL	- Run list contains unmapped elements. Make sure to only pass
  *		  fully mapped runlists to this function.
  *	EIO	- The runlist is corrupt.
@@ -1357,8 +1370,13 @@ int ntfs_mapping_pairs_build(const ntfs_volume *vol, s8 *dst,
 	int i;
 	s8 len_len, lcn_len;
 
-	if (!rl)
+	if (!rl) {
+		if (dst_len < 1)
+			goto size_err;
+		/* Terminator byte. */
+		*dst = 0;
 		return 0;
+	}
 	/*
 	 * @dst_max is used for bounds checking in
 	 * ntfs_write_significant_bytes().
@@ -1435,8 +1453,8 @@ int ntfs_rl_truncate(runlist **arl, const VCN start_vcn)
 	rl = *arl;
 	if (start_vcn < rl->vcn) {
 		// FIXME: Eeek! BUG()
-		fprintf(stderr, "%s(): Eeek! start_vcn lies outside front of "
-				"runlist! Aborting.\n", __FUNCTION__);
+		Dprintf("%s(): Eeek! start_vcn lies outside front of "
+				"runlist!  Aborting.\n", __FUNCTION__);
 		errno = EIO;
 		return -1;
 	}
@@ -1448,14 +1466,14 @@ int ntfs_rl_truncate(runlist **arl, const VCN start_vcn)
 	}
 	if (!rl->length) {
 		// FIXME: Weird, probably a BUG()!
-		fprintf(stderr, "%s(): Weird! Asking to truncate already "
-				"truncated runlist?!? Abort.\n", __FUNCTION__);
+		Dprintf("%s(): Weird!  Asking to truncate already truncated "
+				"runlist?!?  Abort.\n", __FUNCTION__);
 		errno = EIO;
 		return -1;
 	}
 	if (start_vcn < rl->vcn) {
 		// FIXME: Eeek! BUG()
-		fprintf(stderr, "%s(): Eeek! start_vcn < rl->vcn! Aborting.\n",
+		Dprintf("%s(): Eeek!  start_vcn < rl->vcn!  Aborting.\n",
 				__FUNCTION__);
 		errno = EIO;
 		return -1;
@@ -1489,10 +1507,9 @@ int ntfs_rl_truncate(runlist **arl, const VCN start_vcn)
 			*arl = NULL;
 		else {
 			// FIXME: Eeek!
-			fprintf(stderr, "%s(): Eeek! Failed to reallocate "
-					"runlist buffer! Continuing "
-					"regardless and returning success.\n",
-					__FUNCTION__);
+			Dprintf("%s(): Eeek!  Failed to reallocate runlist "
+					"buffer!  Continuing regardless and "
+					"returning success.\n", __FUNCTION__);
 		}
 	}
 	/* Done! */
