@@ -92,8 +92,8 @@ static void __ntfs_volume_release(ntfs_volume *v)
 		if (NDevDirty(dev))
 			dev->d_ops->sync(dev);
 		if (dev->d_ops->close(dev))
-			fprintf(stderr, "%s(): Eeek! Failed to close the "
-					"device. Error: %s\n", __FUNCTION__,
+			Dprintf("%s(): Eeek! Failed to close the device.  "
+					"Error: %s\n", __FUNCTION__,
 					strerror(errno));
 	}
 	if (v->vol_name)
@@ -187,8 +187,7 @@ static int ntfs_mft_load(ntfs_volume *vol)
 		Dputs("Error: failed to allocate buffer for attribute list.");
 		goto error_exit;
 	}
-	l = ntfs_get_attribute_value(vol, vol->mft_ni->mrec, ctx->attr,
-			vol->mft_ni->attr_list);
+	l = ntfs_get_attribute_value(vol, ctx->attr, vol->mft_ni->attr_list);
 	if (!l) {
 		Dputs("Error: failed to get value of $MFT/$ATTRIBUTE_LIST.");
 		goto io_error_exit;
@@ -218,8 +217,8 @@ mft_has_no_attr_list:
 		Dperror("Failed to open ntfs attribute");
 		goto error_exit;
 	}
-	/* Set the number of mft records. */
-	vol->nr_mft_records = vol->mft_na->data_size >>
+	/* Set the number of initialized mft records. */
+	vol->nr_mft_records = vol->mft_na->initialized_size >>
 			vol->mft_record_size_bits;
 	/* Read all extents from the $DATA attribute in $MFT. */
 	ntfs_attr_reinit_search_ctx(ctx);
@@ -398,12 +397,12 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long rwflag)
 {
 	LCN mft_zone_size, mft_lcn;
 	s64 br;
-	const char *OK = "OK";
-	const char *FAILED = "FAILED";
 	ntfs_volume *vol;
 	NTFS_BOOT_SECTOR *bs;
 	int eo;
 #ifdef DEBUG
+	const char *OK = "OK";
+	const char *FAILED = "FAILED";
 	BOOL debug = 1;
 #else
 	BOOL debug = 0;
@@ -586,8 +585,10 @@ error_exit:
 ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 {
 	s64 l;
+#ifdef DEBUG
 	const char *OK = "OK";
 	const char *FAILED = "FAILED";
+#endif
 	ntfs_volume *vol;
 	u8 *m = NULL, *m2 = NULL;
 	ntfs_attr_search_ctx *ctx = NULL;
@@ -595,7 +596,7 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 	ntfs_attr *na;
 	ATTR_RECORD *a;
 	VOLUME_INFORMATION *vinf;
-	uchar_t *vname;
+	ntfschar *vname;
 	int i, j, eo;
 	u32 u;
 
@@ -733,7 +734,7 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 		goto error_exit;
 	}
 	vol->upcase_len = na->data_size >> 1;
-	vol->upcase = (uchar_t*)malloc(na->data_size);
+	vol->upcase = (ntfschar*)malloc(na->data_size);
 	if (!vol->upcase) {
 		Dputs(FAILED);
 		Dputs("Not enough memory to load $UpCase.");
@@ -765,7 +766,7 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 		Dperror("Failed to open inode");
 		goto error_exit;
 	}
-	/* Get an ntfs attribute for $UpCase/$DATA. */
+	/* Get a search context for the $Volume/$VOLUME_INFORMATION lookup. */
 	ctx = ntfs_attr_get_search_ctx(ni, NULL);
 	if (!ctx) {
 		Dputs(FAILED);
@@ -793,7 +794,7 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 	vinf = (VOLUME_INFORMATION*)(le16_to_cpu(a->value_offset) + (char*)a);
 	/* Sanity checks. */
 	if ((char*)vinf + le32_to_cpu(a->value_length) > (char*)ctx->mrec +
-			le16_to_cpu(ctx->mrec->bytes_in_use) ||
+			le32_to_cpu(ctx->mrec->bytes_in_use) ||
 			le16_to_cpu(a->value_offset) + le32_to_cpu(
 			a->value_length) > le32_to_cpu(a->length)) {
 		Dputs(FAILED);
@@ -808,46 +809,70 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long rwflag)
 	/* Do not use le16_to_cpu() macro here as our VOLUME_FLAGS are
 	   defined using cpu_to_le16() macro and hence are consistent. */
 	vol->flags = vinf->flags;
-	/* Find the $VOLUME_NAME attribute. */
+	/*
+	 * Reinitialize the search context for the $Volume/$VOLUME_NAME lookup.
+	 */
 	ntfs_attr_reinit_search_ctx(ctx);
 	if (ntfs_attr_lookup(AT_VOLUME_NAME, AT_UNNAMED, 0, 0, 0, NULL, 0,
 			ctx)) {
-		Dputs(FAILED);
-		Dputs("$VOLUME_NAME attribute not found in $Volume?!?");
-		goto error_exit;
-	}
-	a = ctx->attr;
-	/* Has to be resident. */
-	if (a->non_resident) {
-		Dputs(FAILED);
-		Dputs("Error: Attribute $VOLUME_NAME must be resident!");
-		errno = EIO;
-		goto error_exit;
-	}
-	/* Get a pointer to the value of the attribute. */
-	vname = (uchar_t*)(le16_to_cpu(a->value_offset) + (char*)a);
-	u = le32_to_cpu(a->value_length) / 2;
-	/* Convert Unicode volume name to current locale multibyte format. */
-	vol->vol_name = NULL;
-	if (ntfs_ucstombs(vname, u, &vol->vol_name, 0) == -1) {
-		Dperror("Error: Volume name could not be converted to "
-				"current locale");
-		Dputs("Forcing name into ASCII by replacing non-ASCII "
-				"characters with underscores.");
-		vol->vol_name = malloc(u + 1);
+		if (errno != ENOENT) {
+			Dputs(FAILED);
+			Dputs("Error: Lookup of $VOLUME_NAME attribute in "
+					"$Volume failed.  This probably means "
+					"something is corrupt.  Run chkdsk.");
+			goto error_exit;
+		}
+		/*
+		 * Attribute not present.  This has been seen in the field.
+		 * Treat this the same way as if the attribute was present but
+		 * had zero length.
+		 */
+		vol->vol_name = malloc(1);
 		if (!vol->vol_name) {
 			Dputs(FAILED);
 			Dputs("Error: Unable to allocate memory for volume "
 					"name!");
 			goto error_exit;
 		}
-		for (j = 0; j < (s32)u; j++) {
-			uchar_t uc = le16_to_cpu(vname[j]);
-			if (uc > 0xff)
-				uc = (uchar_t)'_';
-			vol->vol_name[j] = (char)uc;
+		vol->vol_name[0] = '\0';
+	} else {
+		a = ctx->attr;
+		/* Has to be resident. */
+		if (a->non_resident) {
+			Dputs(FAILED);
+			Dputs("Error: Attribute $VOLUME_NAME must be "
+					"resident!");
+			errno = EIO;
+			goto error_exit;
+		}
+		/* Get a pointer to the value of the attribute. */
+		vname = (ntfschar*)(le16_to_cpu(a->value_offset) + (char*)a);
+		u = le32_to_cpu(a->value_length) / 2;
+		/*
+		 * Convert Unicode volume name to current locale multibyte
+		 * format.
+		 */
+		vol->vol_name = NULL;
+		if (ntfs_ucstombs(vname, u, &vol->vol_name, 0) == -1) {
+			Dperror("Error: Volume name could not be converted to "
+					"current locale");
+			Dputs("Forcing name into ASCII by replacing non-ASCII "
+					"characters with underscores.");
+			vol->vol_name = malloc(u + 1);
+			if (!vol->vol_name) {
+				Dputs(FAILED);
+				Dputs("Error: Unable to allocate memory for "
+						"volume name!");
+				goto error_exit;
+			}
+			for (j = 0; j < (s32)u; j++) {
+				ntfschar uc = le16_to_cpu(vname[j]);
+				if (uc > 0xff)
+					uc = (ntfschar)'_';
+				vol->vol_name[j] = (char)uc;
+			}
+			vol->vol_name[u] = '\0';
 		}
-		vol->vol_name[u] = '\0';
 	}
 	Dputs(OK);
 	ntfs_attr_put_search_ctx(ctx);
@@ -1097,7 +1122,8 @@ static int ntfs_mntent_check(const char *file, unsigned long *mnt_flags)
  *
  * On error return -1 with errno set to the error code.
  */
-int ntfs_check_if_mounted(const char *file, unsigned long *mnt_flags)
+int ntfs_check_if_mounted(const char *file __attribute__((unused)),
+		unsigned long *mnt_flags)
 {
 	*mnt_flags = 0;
 #ifdef HAVE_MNTENT_H
@@ -1112,15 +1138,15 @@ int ntfs_check_if_mounted(const char *file, unsigned long *mnt_flags)
  * @vol:	ntfs volume whose version we're interested in.
  *
  * The function checks if the NTFS volume version is known or not.
- * Version 1.1 and 1.2 are used by Windows NT4.
- * Version 2.x is used by Windows 2000 Beta's
+ * Version 1.1 and 1.2 are used by Windows NT3.x and NT4.
+ * Version 2.x is used by Windows 2000 Betas.
  * Version 3.0 is used by Windows 2000.
  * Version 3.1 is used by Windows XP and Windows Server 2003.
  *
  * Return 0 if NTFS version is supported otherwise -1 with errno set.
  *
  * The following error codes are defined:
- *	ENOTSUP   Unknown NTFS versions
+ *	ENOTSUP   Unknown NTFS version
  *	EINVAL	  Invalid argument
  */
 int ntfs_version_is_supported(ntfs_volume *vol)
