@@ -2,8 +2,10 @@
  * ntfsclone - Part of the Linux-NTFS project.
  *
  * Copyright (c) 2003-2004 Szabolcs Szakacsits
+ * Copyright (c) 2004 Anton Altaparmakov
+ * Special image format support copyright (c) 2004 Per Olofsson
  *
- * Clone NTFS data and/or metadata to a sparse file, device or stdout.
+ * Clone NTFS data and/or metadata to a sparse file, image, device or stdout.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,6 +20,7 @@
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <sys/ioctl.h>
 #ifdef HAVE_SYS_VFS_H
 #	include <sys/vfs.h>
 #endif
@@ -41,6 +44,13 @@
 #include "runlist.h"
 #include "utils.h"
 
+#if defined(linux) && defined(_IO) && !defined(BLKGETSIZE)
+#define BLKGETSIZE	_IO(0x12,96)  /* Get device size in 512-byte blocks. */
+#endif
+#if defined(linux) && defined(_IOR) && !defined(BLKGETSIZE64)
+#define BLKGETSIZE64	_IOR(0x12,114,size_t)	/* Get device size in bytes. */
+#endif
+
 static const char *EXEC_NAME = "ntfsclone";
 
 struct {
@@ -52,13 +62,17 @@ struct {
 	int std_out;
 	int blkdev_out;		/* output file is block device */   
 	int metadata_only;
+	int save_image;
+	int restore_image;
 	char *output;
 	char *volume;
+	struct statfs stfs;
 } opt;
 
 struct bitmap {
-	u8 *bm;
 	s64 size;
+	u8 *bm;
+	u8 padding[4];	/* Unused: padding to 64 bit. */
 };
 
 struct progress_bar {
@@ -68,13 +82,12 @@ struct progress_bar {
 	float unit;
 };
 
-struct __ntfs_walk_clusters_ctx {
+typedef struct {
 	ntfs_inode *ni;			/* inode being processed */
 	ntfs_attr_search_ctx *ctx;	/* inode attribute being processed */
 	s64 inuse;			/* number of clusters in use */
-};
+} ntfs_walk_clusters_ctx;
 
-typedef struct __ntfs_walk_clusters_ctx ntfs_walk_clusters_ctx;
 typedef int (ntfs_walk_op)(ntfs_inode *ni, void *data);
 
 struct ntfs_walk_cluster {
@@ -86,6 +99,7 @@ struct ntfs_walk_cluster {
 ntfs_volume *vol = NULL;
 struct bitmap lcn_bitmap;
 
+int fd_in;
 int fd_out;
 FILE *msg_out = NULL;
 
@@ -96,6 +110,19 @@ int wiped_unused_mft = 0;
 int wiped_resident_data = 0;
 int wiped_timestamp_data = 0;
 
+#define IMAGE_MAGIC "\0ntfsclone-image"
+#define IMAGE_MAGIC_SIZE 16
+
+struct {
+	char magic[IMAGE_MAGIC_SIZE];
+	u8 major_ver;
+	u8 minor_ver;
+	u32 cluster_size;
+	s64 device_size;
+	s64 nr_clusters;
+	s64 inuse;
+} __attribute__ ((__packed__)) image_hdr;
+
 #define NTFS_MBYTE (1000 * 1000)
 
 #define ERR_PREFIX   "ERROR"
@@ -111,19 +138,13 @@ int wiped_timestamp_data = 0;
 #define read_all(f, p, n)  io_all((f), (p), (n), 0)
 #define write_all(f, p, n) io_all((f), (p), (n), 1)
 
-/* FIXME: They should be #included but Eprintf conflicts with mkntfs's Eprintf */
-extern int Eprintf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-extern int Vprintf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-extern int Qprintf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-/* FIXME: should be 'static' */
-extern int Printf (const char *format, ...) __attribute__ ((format (printf, 1, 2)));
-
 GEN_PRINTF(Eprintf, stderr,  NULL,         FALSE)
 GEN_PRINTF(Vprintf, msg_out, &opt.verbose, TRUE)
 GEN_PRINTF(Qprintf, msg_out, &opt.quiet,   FALSE)
-GEN_PRINTF(Printf,  msg_out, NULL,         FALSE)
-
+static GEN_PRINTF(Printf,  msg_out, NULL,         FALSE)
 
+static void perr_printf(const char *fmt, ...)
+		__attribute__((format(printf, 1, 2)));
 static void perr_printf(const char *fmt, ...)
 {
 	va_list ap;
@@ -137,6 +158,8 @@ static void perr_printf(const char *fmt, ...)
 	fflush(msg_out);
 }
 
+static void err_printf(const char *fmt, ...)
+		__attribute__((format(printf, 1, 2)));
 static void err_printf(const char *fmt, ...)
 {
 	va_list ap;
@@ -148,6 +171,9 @@ static void err_printf(const char *fmt, ...)
 	fflush(msg_out);
 }
 
+static int err_exit(const char *fmt, ...)
+		__attribute__((noreturn))
+		__attribute__((format(printf, 1, 2)));
 static int err_exit(const char *fmt, ...)
 {
 	va_list ap;
@@ -160,7 +186,9 @@ static int err_exit(const char *fmt, ...)
 	exit(1);
 }
 
-
+static int perr_exit(const char *fmt, ...)
+		__attribute__((noreturn))
+		__attribute__((format(printf, 1, 2)));
 static int perr_exit(const char *fmt, ...)
 {
 	va_list ap;
@@ -176,18 +204,21 @@ static int perr_exit(const char *fmt, ...)
 }
 
 
+static void usage(void)	__attribute__((noreturn));
 static void usage(void)
 {
-	Eprintf("\nUsage: %s [options] device\n"
-		"    Efficiently clone NTFS to a sparse file, device or standard output.\n"
+	Eprintf("\nUsage: %s [options] source\n"
+		"    Efficiently clone NTFS to a sparse file, image, device or standard output.\n"
 		"\n"
-		"    -o FILE --output FILE      Clone NTFS to the non-existent FILE\n"
-		"    -O FILE --overwrite FILE   Clone NTFS to FILE, overwriting if exists\n"
-		"    -m      --metadata         Clone *only* metadata (for NTFS experts)\n"
-		"    -f      --force            Force to progress (DANGEROUS)\n"
-		"    -h      --help             Display this help\n"
+		"    -o, --output FILE      Clone NTFS to the non-existent FILE\n"
+		"    -O, --overwrite FILE   Clone NTFS to FILE, overwriting if exists\n"
+		"    -m, --metadata         Clone *only* metadata (for NTFS experts)\n"
+		"    -s, --save-image       Save to the special image format\n"
+		"    -r, --restore-image    Restore from the special image format\n"
+		"    -f, --force            Force to progress (DANGEROUS)\n"
+		"    -h, --help             Display this help\n"
 #ifdef DEBUG
-		"    -d      --debug            Show debug information\n"
+		"    -d, --debug            Show debug information\n"
 #endif
 		"\n"
 		"    If FILE is '-' then send NTFS data to stdout replacing non used\n"
@@ -200,16 +231,18 @@ static void usage(void)
 
 static void parse_options(int argc, char **argv)
 {
-	static const char *sopt = "-dfhmo:O:";
+	static const char *sopt = "-dfhmo:O:rs";
 	static const struct option lopt[] = {
 #ifdef DEBUG
-		{ "debug",	no_argument,		NULL, 'd' },
+		{ "debug",	   no_argument,		NULL, 'd' },
 #endif
-		{ "force",	no_argument,		NULL, 'f' },
-		{ "help",	no_argument,		NULL, 'h' },
-		{ "metadata",	no_argument,		NULL, 'm' },
-		{ "output",	required_argument,	NULL, 'o' },
-		{ "overwrite",	required_argument,	NULL, 'O' },
+		{ "force",	   no_argument,		NULL, 'f' },
+		{ "help",	   no_argument,		NULL, 'h' },
+		{ "metadata",	   no_argument,		NULL, 'm' },
+		{ "output",	   required_argument,	NULL, 'o' },
+		{ "overwrite",	   required_argument,	NULL, 'O' },
+		{ "restore-image", no_argument,		NULL, 'r' },
+		{ "save-image",	   no_argument,		NULL, 's' },
 		{ NULL, 0, NULL, 0 }
 	};
 
@@ -217,7 +250,7 @@ static void parse_options(int argc, char **argv)
 
 	memset(&opt, 0, sizeof(opt));
 
-	while ((c = getopt_long(argc, argv, sopt, lopt, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, sopt, lopt, NULL)) != (char)-1) {
 		switch (c) {
 		case 1:	/* A non-option argument */
 			if (opt.volume)
@@ -241,7 +274,13 @@ static void parse_options(int argc, char **argv)
 		case 'o':
 			if (opt.output)
 				usage();
-			opt.output = argv[optind-1];
+			opt.output = optarg;
+			break;
+		case 'r':
+			opt.restore_image++;
+			break;
+		case 's':
+			opt.save_image++;
 			break;
 		default:
 			err_printf("Unknown option '%s'.\n", argv[optind-1]);
@@ -262,9 +301,21 @@ static void parse_options(int argc, char **argv)
 		usage();
 	}
 
+	if (opt.metadata_only && opt.save_image)
+		err_exit("Saving only metadata to an image is not "
+			 "supported!\n");
+
+	if (opt.metadata_only && opt.restore_image)
+		err_exit("Restoring only metadata from an image is not "
+			 "supported!\n");
+	
 	if (opt.metadata_only && opt.std_out)
 		err_exit("Cloning only metadata to stdout isn't supported!\n");
 
+	if (opt.save_image && opt.restore_image)
+		err_exit("Saving and restoring an image at the same time "
+			 "is not supported!\n");
+	
 	if (!opt.std_out) {
 		struct stat st;
 		
@@ -340,18 +391,29 @@ static s64 nr_clusters_to_bitmap_byte_size(s64 nr_clusters)
 	return bm_bsize;
 }
 
-static int is_critical_metadata(ntfs_walk_clusters_ctx *image)
+static s64 is_critical_metadata(ntfs_walk_clusters_ctx *image, runlist *rl)
 {
-	s64 inode;
-	
-	inode = image->ni->mft_no;
+	s64 inode = image->ni->mft_no;
 
-	if (inode <= LAST_METADATA_INODE)
+	if (inode <= LAST_METADATA_INODE) {
 		if (inode != FILE_LogFile)
-			return 1;
+			return rl->length;
+
+		if (image->ctx->attr->type == AT_DATA) {
+			/* Save at least the first 8 KiB of FILE_LogFile */
+			s64 s = (s64)8192 - rl->vcn * vol->cluster_size;
+			if (s > 0) {
+				s = rounded_up_division(s, vol->cluster_size);
+				if (rl->length < s)
+					s = rl->length;
+				return s;
+			}
+			return 0;
+		}
+	}
 	
 	if (image->ctx->attr->type != AT_DATA)
-		return 1;
+		return rl->length;
 
 	return 0;
 }
@@ -365,6 +427,8 @@ static int io_all(void *fd, void *buf, int count, int do_write)
 	while (count > 0) {
 		if (do_write)
 			i = write(*(int *)fd, buf, count);
+		else if (opt.restore_image)
+			i = read(*(int *)fd, buf, count);
 		else
 			i = dev->d_ops->read(dev, buf, count);
 		if (i < 0) {
@@ -382,12 +446,29 @@ static int io_all(void *fd, void *buf, int count, int do_write)
 static void copy_cluster(void)
 {
 	char buff[NTFS_MAX_CLUSTER_SIZE]; /* overflow checked at mount time */
+	u32 csize = opt.restore_image ? image_hdr.cluster_size
+		: vol->cluster_size;
 
-	if (read_all(vol->dev, buff, vol->cluster_size) == -1)
+	if (read_all(opt.restore_image ? (void *)&fd_in : vol->dev, buff,
+		     csize) == -1)
 		perr_exit("read_all");
 
-	if (write_all(&fd_out, buff, vol->cluster_size) == -1)
-		perr_exit("write_all");
+	if (opt.save_image) {
+		char cmd = 1;
+		if (write_all(&fd_out, &cmd, sizeof(cmd)) == -1)
+			perr_exit("write_all");
+	}
+	
+	if (write_all(&fd_out, buff, csize) == -1) {
+		int err = errno;
+		perr_printf("Write failed");
+		if (err == EIO && opt.stfs.f_type == 0x517b)
+			Printf("Apparently you tried to clone to a remote "
+			       "Windows computer but they don't\nhave "
+			       "efficient sparse file handling by default. "
+			       "Please try a different method.\n");
+		exit(1);
+	}
 }
 
 static void lseek_to_cluster(s64 lcn)
@@ -399,71 +480,101 @@ static void lseek_to_cluster(s64 lcn)
 	if (vol->dev->d_ops->seek(vol->dev, pos, SEEK_SET) == (off_t)-1)
 		perr_exit("lseek input");
 
-	if (opt.std_out)
+	if (opt.std_out || opt.save_image)
 		return;
 	
 	if (lseek(fd_out, pos, SEEK_SET) == (off_t)-1)
 		perr_exit("lseek output");
 }
 
+static void image_skip_clusters(s64 count)
+{
+	if (opt.save_image && count > 0) {
+		char buff[1 + sizeof(count)];
+
+		buff[0] = 0;
+		memcpy(buff + 1, &count, sizeof(count));
+		
+		if (write_all(&fd_out, buff, sizeof(buff)) == -1)
+			perr_exit("write_all");
+	}
+}
+
 static void dump_clusters(ntfs_walk_clusters_ctx *image, runlist *rl)
 {
-	int i;
+	s64 i, len; /* number of clusters to copy */
 	
-	if (opt.std_out)
+	if (opt.std_out || !opt.metadata_only)
 		return;
 	
-	if (!opt.metadata_only || !is_critical_metadata(image))
+	if (!(len = is_critical_metadata(image, rl)))
 		return;
-
+	
 	lseek_to_cluster(rl->lcn);
 	
 	/* FIXME: this could give pretty suboptimal performance */
-	for (i = 0; i < rl->length; i++)
+	for (i = 0; i < len; i++)
 		copy_cluster();
 }
 
 static void clone_ntfs(u64 nr_clusters)
 {
-	s64 i, pos, count;
+	s64 i, pos, count, last_cl;
 	u8 bm[NTFS_BUF_SIZE];
 	void *buf;
 	u32 csize = vol->cluster_size;
 	u64 p_counter = 0;
 	struct progress_bar progress;
 
-	Printf("Cloning NTFS ...\n");
-	
+	if (opt.save_image)
+		Printf("Saving NTFS to image ...\n");
+	else
+		Printf("Cloning NTFS ...\n");
+
 	if ((buf = calloc(1, csize)) == NULL)
 		perr_exit("dump_to_stdout");
-	
+
 	progress_init(&progress, p_counter, nr_clusters, 100);
+
+	if (opt.save_image) {
+		if (write_all(&fd_out, &image_hdr, sizeof(image_hdr))
+		    == -1)
+			perr_exit("write_all");
+	}
 	
 	pos = 0;
+	last_cl = 0;
 	while (1) {
 		count = ntfs_attr_pread(vol->lcnbmp_na, pos, NTFS_BUF_SIZE, bm);
 		if (count == -1)
 			perr_exit("Couldn't read $Bitmap (pos = %lld)\n", pos);
 
-		if (count == 0)
+		if (count == 0) {
+			image_skip_clusters(pos * 8 - last_cl - 1);
 			return;
+		}
 
 		for (i = 0; i < count; i++, pos++) {
 			s64 cl;  /* current cluster */	  
 
 			for (cl = pos * 8; cl < (pos + 1) * 8; cl++) {
 
-				if (cl > vol->nr_clusters - 1)
+				if (cl > vol->nr_clusters - 1) {
+					image_skip_clusters(cl - last_cl - 1);
 					return;
+				}
 				
 				if (ntfs_bit_get(bm, i * 8 + cl % 8)) {
 					progress_update(&progress, ++p_counter);
 					lseek_to_cluster(cl);
+					image_skip_clusters(cl - last_cl - 1);
+					
 					copy_cluster();
+					last_cl = cl;
 					continue;
 				}
 				
-				if (opt.std_out) {
+				if (opt.std_out && !opt.save_image) {
 					progress_update(&progress, ++p_counter);
 					if (write_all(&fd_out, buf, csize) == -1)
 						perr_exit("write_all");
@@ -473,6 +584,59 @@ static void clone_ntfs(u64 nr_clusters)
 	}
 }
 
+static void write_empty_clusters(s32 csize, s64 count,
+				 struct progress_bar *progress, u64 *p_counter)
+{
+	s64 i;
+	char buff[NTFS_MAX_CLUSTER_SIZE];
+	
+	memset(buff, 0, csize);
+
+	for (i = 0; i < count; i++) {
+		if (write_all(&fd_out, buff, csize) == -1)
+			perr_exit("write_all");
+		progress_update(progress, ++(*p_counter));
+	}
+}
+
+static void restore_image(void)
+{
+	s64 pos = 0, count;
+	s32 csize = image_hdr.cluster_size;
+	char cmd;
+	u64 p_counter = 0;
+	struct progress_bar progress;
+
+	Printf("Restoring NTFS from image ...\n");
+
+	progress_init(&progress, p_counter, opt.std_out ?
+		      image_hdr.nr_clusters : image_hdr.inuse, 100);
+
+	while (pos < image_hdr.nr_clusters) {
+		if (read_all(&fd_in, &cmd, sizeof(cmd)) == -1)
+			perr_exit("read_all");
+
+		if (cmd == 0) {
+			if (read_all(&fd_in, &count, sizeof(count)) == -1)
+				perr_exit("read_all");
+			if (opt.std_out)
+				write_empty_clusters(csize, count,
+						     &progress, &p_counter);
+			else {
+				if (lseek(fd_out, count * csize, SEEK_CUR)
+				    == (off_t)-1)
+					perr_exit("lseek output");
+			}
+			pos += count;
+		} else if (cmd == 1) {
+			copy_cluster();
+			pos++;
+			progress_update(&progress, ++p_counter);
+		} else
+			err_exit("Invalid command code in image\n");
+	}
+}
+
 #define WIPE_TIMESTAMPS(atype, attr)				\
 do {								\
 	atype *ats;						\
@@ -506,10 +670,10 @@ static void wipe_resident_data(ntfs_walk_clusters_ctx *image)
 	ATTR_RECORD *a;
 	u32 i;
 	int n = 0;
-	char *p;
+	u8 *p;
 
 	a = image->ctx->attr;
-	p = (char *)a + a->value_offset;
+	p = (u8*)a + le16_to_cpu(a->value_offset);
 	
 	if (image->ni->mft_no <= LAST_METADATA_INODE)
 		return;
@@ -517,7 +681,7 @@ static void wipe_resident_data(ntfs_walk_clusters_ctx *image)
 	if (a->type != AT_DATA)
 		return;
 	
-	for (i = 0; i < a->value_length; i++) {
+	for (i = 0; i < le32_to_cpu(a->value_length); i++) {
 		if (p[i]) {
 			p[i] = 0;
 			n++;
@@ -559,7 +723,8 @@ static void walk_runs(struct ntfs_walk_cluster *walk)
 		if (lcn < 0 || lcn_length < 0)
 			err_exit("Corrupt runlist in inode %lld attr %x LCN "
 				 "%llx length %llx\n", ctx->ntfs_ino->mft_no,
-				 le32_to_cpu (a->type), lcn, lcn_length);
+				 (unsigned int)le32_to_cpu(a->type), lcn,
+				 lcn_length);
 		
 		if (!wipe)
 			dump_clusters(walk->image, rl + i);
@@ -567,9 +732,9 @@ static void walk_runs(struct ntfs_walk_cluster *walk)
 		for (j = 0; j < lcn_length; j++) {
 			u64 k = (u64)lcn + j;
 			if (ntfs_bit_get_and_set(lcn_bitmap.bm, k, 1))
-				err_exit("Cluster %lu referenced twice!\n"
+				err_exit("Cluster %llu referenced twice!\n"
 					 "You didn't shutdown your Windows"
-					 "properly?\n", k);
+					 "properly?\n", (unsigned long long)k);
 		}
 
 		walk->image->inuse += lcn_length;
@@ -680,8 +845,9 @@ static void wipe_unused_mft_data(ntfs_inode *ni)
 	if (ni->mft_no <= LAST_METADATA_INODE)
 		return;
 	
-	unused = m->bytes_allocated - m->bytes_in_use;
-	wiped_unused_mft_data += wipe_data((char *)m, m->bytes_in_use, unused);
+	unused = le32_to_cpu(m->bytes_allocated) - le32_to_cpu(m->bytes_in_use);
+	wiped_unused_mft_data += wipe_data((char *)m,
+			le32_to_cpu(m->bytes_in_use), unused);
 }
 
 static void wipe_unused_mft(ntfs_inode *ni)
@@ -693,8 +859,7 @@ static void wipe_unused_mft(ntfs_inode *ni)
 	if (ni->mft_no <= LAST_METADATA_INODE)
 		return;
 	
-	/* MFT_RECORD doesn't have the XP specific 6 bytes, so add it */
-	unused = m->bytes_in_use - (sizeof(MFT_RECORD) + 6);
+	unused = le32_to_cpu(m->bytes_in_use) - sizeof(MFT_RECORD);
 	wiped_unused_mft += wipe_data((char *)m, sizeof(MFT_RECORD), unused);
 }
 
@@ -763,7 +928,7 @@ static int walk_clusters(ntfs_volume *volume, struct ntfs_walk_cluster *walk)
 		if (wipe)
 			nr_used_mft_records++;
 
-		if ((ni->mrec->base_mft_record) != 0)
+		if (ni->mrec->base_mft_record)
 			goto out;
 
 		walk->image->ni = ni;
@@ -824,12 +989,12 @@ static void print_volume_size(const char *str, s64 bytes)
 }
 
 
-static void print_disk_usage(ntfs_walk_clusters_ctx *image)
+static void print_disk_usage(u32 cluster_size, s64 nr_clusters, s64 inuse)
 {
 	s64 total, used;
 
-	total = vol->nr_clusters * vol->cluster_size;
-	used = image->inuse * vol->cluster_size;
+	total = nr_clusters * cluster_size;
+	used = inuse * cluster_size;
 
 	Printf("Space in use       : %lld MB (%.1f%%)   ",
 			(long long)rounded_up_division(used, NTFS_MBYTE),
@@ -838,6 +1003,37 @@ static void print_disk_usage(ntfs_walk_clusters_ctx *image)
 	Printf("\n");
 }
 
+static void print_image_info(void)
+{
+	Printf("NTFS volume version: %d.%d\n",
+	       image_hdr.major_ver, image_hdr.minor_ver);
+	Printf("Cluster size       : %u bytes\n",
+	       (unsigned int)image_hdr.cluster_size);
+	print_volume_size("Image volume size  ",
+			  image_hdr.nr_clusters * image_hdr.cluster_size);
+	Printf("Image device size  : %lld bytes\n", image_hdr.device_size);
+	print_disk_usage(image_hdr.cluster_size,
+			 image_hdr.nr_clusters,
+			 image_hdr.inuse);
+}
+
+static void check_if_mounted(const char *device, unsigned long new_mntflag)
+{
+	unsigned long mntflag;
+	
+	if (ntfs_check_if_mounted(device, &mntflag))
+		perr_exit("Failed to check '%s' mount state", device);
+
+	if (mntflag & NTFS_MF_MOUNTED) {
+		if (!(mntflag & NTFS_MF_READONLY))
+			err_exit("Device %s is mounted read-write. "
+				 "You must 'umount' it first.\n", device);
+		if (!new_mntflag)
+			err_exit("Device %s is mounted. "
+				 "You must 'umount' it first.\n", device);
+	}
+}
+
 /**
  * First perform some checks to determine if the volume is already mounted, or
  * is dirty (Windows wasn't shutdown properly).  If everything is OK, then mount
@@ -845,20 +1041,8 @@ static void print_disk_usage(ntfs_walk_clusters_ctx *image)
  */
 static void mount_volume(unsigned long new_mntflag)
 {
-	unsigned long mntflag;
+	check_if_mounted(opt.volume, new_mntflag);
 	
-	if (ntfs_check_if_mounted(opt.volume, &mntflag))
-		perr_exit("Failed to check '%s' mount state", opt.volume);
-
-	if (mntflag & NTFS_MF_MOUNTED) {
-		if (!(mntflag & NTFS_MF_READONLY))
-			err_exit("Device %s is mounted read-write. "
-				 "You must 'umount' it first.\n", opt.volume);
-		if (!new_mntflag)
-			err_exit("Device %s is mounted. "
-				 "You must 'umount' it first.\n", opt.volume);
-	}
-
 	if (!(vol = ntfs_mount(opt.volume, new_mntflag))) {
 
 		int err = errno;
@@ -879,13 +1063,15 @@ static void mount_volume(unsigned long new_mntflag)
 				 "please try again (or see -f option).\n");
 	
 	if (NTFS_MAX_CLUSTER_SIZE < vol->cluster_size)
-		err_exit("Cluster size %u is too large!\n", vol->cluster_size);
+		err_exit("Cluster size %u is too large!\n",
+				(unsigned int)vol->cluster_size);
 
 	Printf("NTFS volume version: %d.%d\n", vol->major_ver, vol->minor_ver);
 	if (ntfs_version_is_supported(vol))
 		perr_exit("Unknown NTFS version");
 
-	Printf("Cluster size       : %u bytes\n", vol->cluster_size);
+	Printf("Cluster size       : %u bytes\n",
+			(unsigned int)vol->cluster_size);
 	print_volume_size("Current volume size",
 			  volume_size(vol, vol->nr_clusters));
 }
@@ -911,7 +1097,7 @@ static s64 device_size_get(int fd)
 			Dprintf("BLKGETSIZE64 nr bytes = %llu (0x%llx)\n",
 					(unsigned long long)size,
 					(unsigned long long)size);
-			return (s64)size / block_size;
+			return (s64)size;
 		}
 	}
 #endif
@@ -961,10 +1147,106 @@ static void fsync_clone(int fd)
 		perr_exit("fsync");
 }
 
+static void set_filesize(s64 filesize)
+{
+	if (fstatfs(fd_out, &opt.stfs) == -1)
+		Printf("WARNING: Couldn't get filesystem type: "
+		       "%s\n", strerror(errno));
+	else if (opt.stfs.f_type == 0x52654973)
+		Printf("WARNING: You're using ReiserFS, it has very poor "
+		       "performance creating\nlarge sparse files. The next "
+		       "operation might take a very long time!\n"
+		       "Creating sparse output file ...\n");
+	else if (opt.stfs.f_type == 0x517b)
+		Printf("WARNING: You're using SMBFS and if the remote share "
+		       "isn't Samba but a Windows\ncomputer then the clone "
+		       "operation will be very inefficient and may fail!\n");
+
+	if (ftruncate(fd_out, filesize) == -1) {
+		int err = errno;
+		perr_printf("ftruncate failed for file '%s'", opt.output);
+		if (err == E2BIG) {
+			Printf("Your system or the destination filesystem "
+			       "doesn't support large files.\n");
+			if (opt.stfs.f_type == 0x517b) {
+				Printf("SMBFS needs minimum Linux kernel "
+				       "version 2.4.25 and\n the 'lfs' option"
+				       "\nfor smbmount to have large "
+				       "file support.\n");
+			}
+		}
+		exit(1);
+	}
+}
+
+static s64 open_image(void)
+{
+	if (strcmp(opt.volume, "-") == 0) {
+		if ((fd_in = fileno(stdin)) == -1) 
+			perr_exit("fileno for stdout failed");
+	} else {
+		if ((fd_in = open(opt.volume, O_RDONLY)) == -1)
+			perr_exit("failed to open image");
+	}
+
+	if (read_all(&fd_in, &image_hdr, sizeof(image_hdr)) == -1)
+		perr_exit("read_all");
+
+	if (memcmp(image_hdr.magic, IMAGE_MAGIC, IMAGE_MAGIC_SIZE) != 0)
+		err_exit("Input file is not an image! (invalid magic)\n");
+
+	return image_hdr.device_size;
+}
+
+static s64 open_volume(void)
+{
+	s64 device_size;
+	
+	mount_volume(MS_RDONLY);
+	
+	device_size = ntfs_device_size_get(vol->dev, 1);
+	if (device_size <= 0)
+		err_exit("Couldn't get device size (%lld)!\n", device_size);
+	
+	print_volume_size("Current device size", device_size);
+	
+	if (device_size < vol->nr_clusters * vol->cluster_size)
+		err_exit("Current NTFS volume size is bigger than the device "
+			 "size (%lld)!\nCorrupt partition table or incorrect "
+			 "device partitioning?\n", device_size);
+
+	return device_size;
+}
+
+static void initialise_image_hdr(s64 device_size, s64 inuse)
+{
+	memcpy(image_hdr.magic, IMAGE_MAGIC, IMAGE_MAGIC_SIZE);
+	image_hdr.major_ver = vol->major_ver;
+	image_hdr.minor_ver = vol->minor_ver;
+	image_hdr.cluster_size = vol->cluster_size;
+	image_hdr.device_size = device_size;
+	image_hdr.nr_clusters = vol->nr_clusters;
+	image_hdr.inuse = inuse;
+}
+
+static void check_output_device(s64 input_size)
+{
+	if (opt.blkdev_out) {
+		s64 dest_size = device_size_get(fd_out);
+		if (dest_size < input_size)
+			err_exit("Output device is too small (%lld) to fit the "
+				 "NTFS image (%lld).\n", dest_size, input_size);
+		
+		check_if_mounted(opt.output, 0);
+	} else
+		set_filesize(input_size);
+}
+
 int main(int argc, char **argv)
 {
 	ntfs_walk_clusters_ctx image;
-	s64 device_size;        /* in bytes */
+	s64 device_size;        /* input device size in bytes */
+	s64 ntfs_size;
 	int wiped_total = 0;
 
 	/* print to stderr, stdout can be an NTFS image ... */
@@ -975,24 +1257,21 @@ int main(int argc, char **argv)
 	
 	utils_set_locale();
 
-	mount_volume(MS_RDONLY);
-	
-	device_size = ntfs_device_size_get(vol->dev, 1);
-	if (device_size <= 0)
-		err_exit("Couldn't get device size (%lld)!\n", device_size);
-
-	print_volume_size("Current device size", device_size);
-
-	if (device_size < vol->nr_clusters * vol->cluster_size)
-		err_exit("Current NTFS volume size is bigger than the device "
-			 "size (%lld)!\nCorrupt partition table or incorrect "
-			 "device partitioning?\n", device_size);
+	if (opt.restore_image) {
+		device_size = open_image();
+		ntfs_size = image_hdr.nr_clusters * image_hdr.cluster_size;
+	} else {
+		device_size = open_volume();
+		ntfs_size = vol->nr_clusters * vol->cluster_size;
+	}
+	ntfs_size += 512; /* add backup boot sector */
 
 	if (opt.std_out) {
 	       if ((fd_out = fileno(stdout)) == -1) 
 		       perr_exit("fileno for stdout failed");
 	} else {
-		int flags = O_WRONLY;
+	        /* device_size_get() might need to read() */
+		int flags = O_RDWR;
 		
 		if (!opt.blkdev_out) {
 			flags |= O_CREAT | O_TRUNC;
@@ -1001,46 +1280,37 @@ int main(int argc, char **argv)
 		}
 
 		if ((fd_out = open(opt.output, flags, S_IRWXU)) == -1) 
-			perr_exit("opening file '%s' failed", opt.output);
-	
-		if (!opt.blkdev_out) {
-			struct statfs stfs;
-			
-			if (fstatfs(fd_out, &stfs) == -1)
-				Printf("Couldn't get Linux filesystem type: "
-				       "%s\n", strerror(errno));
-			else if (stfs.f_type == 0x52654973) {
-				Printf("WARNING: You're using ReiserFS, it has "
-				       "very poor performance creating\nlarge "
-				       "sparse files. The next operation "
-				       "might take a very long time!\n"
-				       "Creating sparse output file ...\n");
-			}
-			if (ftruncate(fd_out, device_size) == -1)
-				perr_exit("ftruncate failed for file '%s'", 
-					  opt.output);
-		} else {
-			s64 dest_size = device_size_get(fd_out);
-			if (dest_size < device_size)
-				err_exit("Output device size (%lld) is too small"
-				    " to fit an NTFS clone\n", dest_size);
-		}
+			perr_exit("Opening file '%s' failed", opt.output);
+
+		if (!opt.save_image)
+			check_output_device(ntfs_size);
 	}
 
+	if (opt.restore_image) {
+		print_image_info();
+		restore_image();
+		fsync_clone(fd_out);
+		exit(0);
+	}
+	
 	setup_lcn_bitmap();
 	memset(&image, 0, sizeof(image));
 	backup_clusters.image = &image; 
 
 	walk_clusters(vol, &backup_clusters);
 	compare_bitmaps(&lcn_bitmap);
-	print_disk_usage(&image);
+	print_disk_usage(vol->cluster_size, vol->nr_clusters, image.inuse);
 	
 	free(lcn_bitmap.bm);
+
+	if (opt.save_image)
+		initialise_image_hdr(device_size, image.inuse);
 	
 	/* FIXME: save backup boot sector */
 
 	if (opt.std_out || !opt.metadata_only) {
-		s64 nr_clusters = opt.std_out ? vol->nr_clusters : image.inuse;
+		s64 nr_clusters = (opt.std_out && !opt.save_image)
+			? vol->nr_clusters : image.inuse;
 		
 		clone_ntfs(nr_clusters);
 		fsync_clone(fd_out);
@@ -1078,4 +1348,3 @@ int main(int argc, char **argv)
 	fsync_clone(fd_out);
 	exit(0);
 }
-
