@@ -71,7 +71,7 @@ BOOL success = FALSE;
 char *dev_name;
 s64 inode;
 u32 attr_type;
-uchar_t *attr_name = NULL;
+ntfschar *attr_name = NULL;
 u32 attr_name_len;
 s64 new_len;
 
@@ -95,7 +95,8 @@ struct {
 /**
  * Dprintf - debugging output (-vv); overriden by quiet (-q)
  */
-void Dprintf(const char *fmt, ...)
+static void Dprintf(const char *fmt, ...) __attribute__((format(printf, 1, 2)));
+static void Dprintf(const char *fmt, ...)
 {
 	va_list ap;
 
@@ -110,7 +111,7 @@ void Dprintf(const char *fmt, ...)
 /**
  * Eprintf - error output; ignores quiet (-q)
  */
-void Eprintf(const char *fmt, ...)
+int Eprintf(const char *fmt, ...)
 {
 	va_list ap;
 
@@ -118,6 +119,7 @@ void Eprintf(const char *fmt, ...)
 	va_start(ap, fmt);
 	vfprintf(stderr, fmt, ap);
 	va_end(ap);
+	return 0;
 }
 
 /* Generate code for Vprintf() function: Verbose output (-v). */
@@ -129,7 +131,10 @@ GEN_PRINTF(Qprintf, stdout, &opts.quiet,   FALSE)
 /**
  * err_exit - error output and terminate; ignores quiet (-q)
  */
-void err_exit(const char *fmt, ...)
+static void err_exit(const char *fmt, ...)
+		__attribute__((noreturn))
+		__attribute__((format(printf, 1, 2)));
+static void err_exit(const char *fmt, ...)
 {
 	va_list ap;
 
@@ -144,7 +149,7 @@ void err_exit(const char *fmt, ...)
 /**
  * copyright - print copyright statements
  */
-void copyright(void)
+static void copyright(void)
 {
 	fprintf(stderr, "Copyright (c) 2002-2003 Anton Altaparmakov\n"
 			"Copyright (c) 2003 Richard Russon\n"
@@ -155,7 +160,7 @@ void copyright(void)
 /**
  * license - print licese statement
  */
-void license(void)
+static void license(void)
 {
 	fprintf(stderr, "%s", ntfs_gpl);
 }
@@ -188,7 +193,7 @@ void usage (void)
 /**
  * parse_options
  */
-void parse_options(int argc, char *argv[])
+static void parse_options(int argc, char *argv[])
 {
 	long long ll;
 	char *s, *s2;
@@ -263,7 +268,7 @@ void parse_options(int argc, char *argv[])
 		if (optind != argc) {
 			/* Convert the string to little endian Unicode. */
 			attr_name_len = ntfs_mbstoucs(s, &attr_name, 0);
-			if (attr_name_len < 0)
+			if ((int)attr_name_len < 0)
 				err_exit("Invalid attribute name \"%s\": %s\n",
 						s, strerror(errno));
 
@@ -278,12 +283,13 @@ void parse_options(int argc, char *argv[])
 			attr_name_len = 0;
 		}
 	}
-	Dprintf("attribute type = 0x%x\n", attr_type);
+	Dprintf("attribute type = 0x%x\n", (unsigned int)attr_type);
 	if (attr_name == AT_UNNAMED)
 		Dprintf("attribute name = \"\" (UNNAMED)\n");
 	else
-		Dprintf("attribute name = \"%s\" (length %i Unicode "
-				"characters)\n", s2, attr_name_len);
+		Dprintf("attribute name = \"%s\" (length %u Unicode "
+				"characters)\n", s2,
+				(unsigned int)attr_name_len);
 
 	/* Get the new length. */
 	ll = strtoll(s, &s2, 0);
@@ -303,9 +309,9 @@ void parse_options(int argc, char *argv[])
  * terminating null byte. If a unicode character was encountered which could
  * not be converted -1 is returned.
  */
-int ucstos(char *dest, const uchar_t *src, int maxlen)
+static int ucstos(char *dest, const ntfschar *src, int maxlen)
 {
-	uchar_t u;
+	ntfschar u;
 	int i;
 
 	/* Need one byte for null terminator. */
@@ -325,7 +331,7 @@ int ucstos(char *dest, const uchar_t *src, int maxlen)
 /**
  * dump_resident_attr_val
  */
-void dump_resident_attr_val(ATTR_TYPES type, char *val, u32 val_len)
+static void dump_resident_attr_val(ATTR_TYPES type, char *val, u32 val_len)
 {
 	const char *don_t_know = "Don't know what to do with this attribute "
 			"type yet.";
@@ -357,13 +363,13 @@ void dump_resident_attr_val(ATTR_TYPES type, char *val, u32 val_len)
 		printf("%s\n", todo);
 		return;
 	case AT_VOLUME_NAME:
-		printf("Volume name length = %i\n", val_len);
+		printf("Volume name length = %u\n", (unsigned int)val_len);
 		if (val_len) {
 			buf = calloc(1, val_len);
 			if (!buf)
 				err_exit("Failed to allocate internal buffer: "
 						"%s\n", strerror(errno));
-			i = ucstos(buf, (uchar_t*)val, val_len);
+			i = ucstos(buf, (ntfschar*)val, val_len);
 			if (i == -1)
 				printf("Volume name contains non-displayable "
 						"Unicode characters.\n");
@@ -466,14 +472,14 @@ void dump_resident_attr_val(ATTR_TYPES type, char *val, u32 val_len)
 		printf("Cannot display unknown %s defined attribute type 0x%x"
 				".\n", u >=
 				le32_to_cpu(AT_FIRST_USER_DEFINED_ATTRIBUTE) ?
-				"user" : "system", u);
+				"user" : "system", (unsigned int)u);
 	}
 }
 
 /**
  * dump_resident_attr
  */
-void dump_resident_attr(ATTR_RECORD *a)
+static void dump_resident_attr(ATTR_RECORD *a)
 {
 	int i;
 
@@ -496,7 +502,7 @@ void dump_resident_attr(ATTR_RECORD *a)
 /**
  * dump_mapping_pairs_array
  */
-void dump_mapping_pairs_array(char *b, unsigned int max_len)
+static void dump_mapping_pairs_array(char *b, unsigned int max_len)
 {
 	// TODO
 	return;
@@ -505,7 +511,7 @@ void dump_mapping_pairs_array(char *b, unsigned int max_len)
 /**
  * dump_non_resident_attr
  */
-void dump_non_resident_attr(ATTR_RECORD *a)
+static void dump_non_resident_attr(ATTR_RECORD *a)
 {
 	s64 l;
 	int i;
@@ -544,7 +550,7 @@ void dump_non_resident_attr(ATTR_RECORD *a)
 /**
  * dump_attr_record
  */
-void dump_attr_record(MFT_RECORD *m, ATTR_RECORD *a)
+static void dump_attr_record(MFT_RECORD *m, ATTR_RECORD *a)
 {
 	unsigned int u;
 	char s[0x200];
@@ -553,7 +559,8 @@ void dump_attr_record(MFT_RECORD *m, ATTR_RECORD *a)
 	printf("-- Beginning dump of attribute record at offset 0x%x. --\n",
 			(unsigned)((u8*)a - (u8*)m));
 	if (a->type == AT_END) {
-		printf("Attribute type = 0x%x ($END)\n", le32_to_cpu(AT_END));
+		printf("Attribute type = 0x%x ($END)\n",
+				(unsigned int)le32_to_cpu(AT_END));
 		u = le32_to_cpu(a->length);
 		printf("Length of resident part = %u (0x%x)\n", u, u);
 		return;
@@ -584,7 +591,7 @@ void dump_attr_record(MFT_RECORD *m, ATTR_RECORD *a)
 			cpu_to_le16(a->name_offset));
 	u = a->flags;
 	if (a->name_length) {
-		if (ucstos(s, (uchar_t*)((char*)a +
+		if (ucstos(s, (ntfschar*)((char*)a +
 				cpu_to_le16(a->name_offset)),
 				min((int)sizeof(s),
 						a->name_length + 1)) == -1) {
@@ -641,7 +648,7 @@ void dump_attr_record(MFT_RECORD *m, ATTR_RECORD *a)
 /**
  * dump_mft_record
  */
-void dump_mft_record(MFT_RECORD *m)
+static void dump_mft_record(MFT_RECORD *m)
 {
 	ATTR_RECORD *a;
 	unsigned int u;
@@ -694,7 +701,7 @@ void dump_mft_record(MFT_RECORD *m)
 /**
  * ntfstruncate_exit
  */
-void ntfstruncate_exit(void)
+static void ntfstruncate_exit(void)
 {
 	int err;
 
@@ -780,8 +787,8 @@ int main(int argc, char **argv)
 	/* Open the specified attribute. */
 	na = ntfs_attr_open(ni, attr_type, attr_name, attr_name_len);
 	if (!na)
-		err_exit("Failed to open attribute 0x%x: %s\n", attr_type,
-				strerror(errno));
+		err_exit("Failed to open attribute 0x%x: %s\n",
+				(unsigned int)attr_type, strerror(errno));
 
 	if (!opts.quiet && opts.verbose > 1) {
 		Dprintf("Dumping mft record before calling "
@@ -792,8 +799,8 @@ int main(int argc, char **argv)
 	/* Truncate the attribute. */
 	err = ntfs_attr_truncate(na, new_len);
 	if (err)
-		err_exit("Failed to truncate attribute 0x%x: %s\n", attr_type,
-				strerror(errno));
+		err_exit("Failed to truncate attribute 0x%x: %s\n",
+				(unsigned int)attr_type, strerror(errno));
 
 	if (!opts.quiet && opts.verbose > 1) {
 		Dprintf("Dumping mft record after calling "
