@@ -1,7 +1,8 @@
 /**
  * ntfsundelete - Part of the Linux-NTFS project.
  *
- * Copyright (c) 2002-2003 Richard Russon
+ * Copyright (c) 2002-2004 Richard Russon
+ * Copyright (c) 2004 Holger Ohmacht
  *
  * This utility will recover deleted files from an NTFS volume.
  *
@@ -54,17 +55,139 @@
 
 static const char *EXEC_NAME = "ntfsundelete";
 static const char *MFTFILE   = "mft";
+#ifdef DEBUG
 static const char *UNNAMED   = "<unnamed>";
-static       char *NONE      = "<none>";
-static       char *UNKNOWN   = "unknown";
+#endif
+static const char *NONE      = "<none>";
+static const char *UNKNOWN   = "unknown";
 static struct options opts;
 
+typedef struct
+{
+	u32 begin;
+	u32 end;
+} range;
+
+static range *ranges;
+static long  nr_entries;
+
 GEN_PRINTF (Eprintf, stderr, NULL,          FALSE)
 GEN_PRINTF (Vprintf, stdout, &opts.verbose, TRUE)
 GEN_PRINTF (Qprintf, stdout, &opts.quiet,   FALSE)
 
 #define _(S)	gettext(S)
 
+/**
+ * parse_inode_arg - parses the inode expression
+ *
+ * Parses the optarg after parameter -u for valid ranges
+ *
+ * Return: Number of correct inode specifications or -1 for error
+ */
+static int parse_inode_arg (void)
+{
+	int p;
+	u32 imax;
+	u32 range_begin;
+	u32 range_end;
+	u32 range_temp;
+	u32 inode;
+	char *opt_arg_ptr;
+	char *opt_arg_temp;
+	char *opt_arg_end1;
+	char *opt_arg_end2;
+
+	/* init variables */
+	p = strlen (optarg);
+	imax = p;
+	opt_arg_ptr = optarg;
+	opt_arg_end1 = optarg;
+	opt_arg_end2 = &(optarg[p]);
+	nr_entries = 0;
+
+	/* alloc mem for range table */
+	ranges = (range *) malloc ((p + 1) * sizeof (range));
+	if (ranges == NULL)
+	{
+		printf ("ERROR: Couldn't alloc mem for parsing inodes!\n");
+		return (-1);
+	}
+
+	/* loop */
+	while ((opt_arg_end1 != opt_arg_end2) && (p > 0))
+	{
+		inode = strtoul (opt_arg_ptr, &opt_arg_end1, 0);
+		p--;
+
+		/* invalid char at begin */
+		if ((opt_arg_ptr == opt_arg_end1) || (opt_arg_ptr == opt_arg_end2))
+		{
+			printf ("ERROR: Invalid Number: %s\n", opt_arg_ptr);
+			return (-1);
+		}
+
+		/* RANGE - Check for range */
+		if (opt_arg_end1[0] == '-')
+		{
+			/* get range end */
+			opt_arg_temp = opt_arg_end1;
+			opt_arg_end1 = & (opt_arg_temp[1]);
+			if (opt_arg_temp >= opt_arg_end2)
+			{
+				printf ("ERROR: Missing range end!\n");
+				return (-1);
+			}
+			range_begin = inode;
+
+			/* get count */
+			range_end = strtoul (opt_arg_end1, &opt_arg_temp, 0);
+			if (opt_arg_temp == opt_arg_end1)
+			{
+				printf ("ERROR: Invalid Number: %s\n", opt_arg_temp);
+				return (-1);
+			}
+
+			/* check for correct values */
+			if (range_begin > range_end)
+			{
+				range_temp = range_end;
+				range_end = range_begin;
+				range_begin = range_temp;
+			}
+
+			/* put into struct */
+			ranges[nr_entries].begin = range_begin;
+			ranges[nr_entries].end = range_end;
+			nr_entries++;
+
+			/* Last check */
+			opt_arg_ptr = & (opt_arg_temp[1]);
+			if (opt_arg_ptr >= opt_arg_end2)
+				break;
+		} else if (opt_arg_end1[0] == ',') {
+			/* SINGLE VALUE, BUT CONTINUING */
+			/* put inode into range list */
+			ranges[nr_entries].begin = inode;
+			ranges[nr_entries].end = inode;
+			nr_entries++;
+
+			/* Next inode */
+			opt_arg_ptr = & (opt_arg_end1[1]);
+			if (opt_arg_ptr >= opt_arg_end2)
+			{
+				printf ("ERROR: Missing new value at end of input!\n");
+				return (-1);
+			}
+			continue;
+		} else { /* SINGLE VALUE, END */
+			ranges[nr_entries].begin = inode;
+			ranges[nr_entries].end = inode;
+			nr_entries++;
+		}
+	}
+	return (nr_entries);
+}
+
 /**
  * version - Print version information about the program
  *
@@ -72,7 +195,7 @@ GEN_PRINTF (Qprintf, stdout, &opts.quiet,   FALSE)
  *
  * Return:  none
  */
-void version (void)
+static void version (void)
 {
 	printf ("\n%s v%s - Recover deleted files from an NTFS Volume.\n\n",
 		EXEC_NAME, VERSION);
@@ -88,7 +211,7 @@ void version (void)
  *
  * Return:  none
  */
-void usage (void)
+static void usage (void)
 {
 	printf ("\nUsage: %s [options] device\n"
 		"    -s          --scan             Scan for files (default)\n"
@@ -98,10 +221,13 @@ void usage (void)
 		"    -S range    --size range       Match files of this size\n"
 		"    -t since    --time since       Last referenced since this time\n"
 		"\n"
-		"    -u num      --undelete num     Undelete inode\n"
+		"    -u nums     --undelete nums    Undelete inodes\n"
 		"    -o file     --output file      Save with this filename\n"
+		"    -O          --optimistic       Undelete in-use clusters as well\n"
 		"    -d dir      --destination dir  Destination directory\n"
 		"    -b num      --byte num         Fill missing parts with this byte\n"
+		"    -T          --truncate         Truncate 100%% recoverable file to exact size.\n"
+		"    -P          --parent           Show parent directory\n"
 		"\n"
 		"    -c range    --copy range       Write a range of MFT records to a file\n"
 		"\n"
@@ -135,7 +261,7 @@ void usage (void)
  * Return:  1, Success, the string was transformed
  *	    0, An error occurred
  */
-int transform (const char *pattern, char **regex)
+static int transform (const char *pattern, char **regex)
 {
 	char *result;
 	int length, i, j;
@@ -210,9 +336,10 @@ int transform (const char *pattern, char **regex)
  * Return:  1  Success
  *	    0  Error, the string was malformed
  */
-int parse_time (const char *value, time_t *since)
+static int parse_time (const char *value, time_t *since)
 {
-	time_t result, now;
+	long long result;
+	time_t now;
 	char *suffix = NULL;
 
 	if (!value || !since)
@@ -266,9 +393,9 @@ int parse_time (const char *value, time_t *since)
  * Return:  1 Success
  *	    0 Error, one or more problems
  */
-int parse_options (int argc, char *argv[])
+static int parse_options (int argc, char *argv[])
 {
-	static const char *sopt = "-b:Cc:d:fh?m:o:p:sS:t:u:qvV";
+	static const char *sopt = "-b:Cc:d:fh?m:o:OPp:sS:t:Tu:qvV";
 	static const struct option lopt[] = {
 		{ "byte",	 required_argument,	NULL, 'b' },
 		{ "case",	 no_argument,		NULL, 'C' },
@@ -278,10 +405,13 @@ int parse_options (int argc, char *argv[])
 		{ "help",	 no_argument,		NULL, 'h' },
 		{ "match",	 required_argument,	NULL, 'm' },
 		{ "output",	 required_argument,	NULL, 'o' },
+		{ "optimistic",  no_argument,           NULL, 'O' },
 		{ "percentage",  required_argument,	NULL, 'p' },
 		{ "scan",	 no_argument,		NULL, 's' },
 		{ "size",	 required_argument,	NULL, 'S' },
+		{ "parent",	 no_argument,		NULL, 'P' },
 		{ "time",	 required_argument,	NULL, 't' },
+		{ "truncate",	 no_argument,		NULL, 'T' },
 		{ "undelete",	 required_argument,	NULL, 'u' },
 		{ "quiet",	 no_argument,		NULL, 'q' },
 		{ "verbose",	 no_argument,		NULL, 'v' },
@@ -302,7 +432,7 @@ int parse_options (int argc, char *argv[])
 	opts.percent  = -1;
 	opts.fillbyte = -1;
 
-	while ((c = getopt_long (argc, argv, sopt, lopt, NULL)) != -1) {
+	while ((c = getopt_long (argc, argv, sopt, lopt, NULL)) != (char)-1) {
 		switch (c) {
 		case 1:	/* A non-option argument */
 			if (!opts.device) {
@@ -313,7 +443,7 @@ int parse_options (int argc, char *argv[])
 			}
 			break;
 		case 'b':
-			if (opts.fillbyte == -1) {
+			if (opts.fillbyte == (char)-1) {
 				end = NULL;
 				opts.fillbyte = strtol (optarg, &end, 0);
 				if (end && *end)
@@ -327,7 +457,7 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'c':
 			if (opts.mode == MODE_NONE) {
-				if (!utils_parse_range (argv[optind-1],
+				if (!utils_parse_range (optarg,
 				    &opts.mft_begin, &opts.mft_end, TRUE))
 					err++;
 				opts.mode = MODE_COPY;
@@ -337,7 +467,7 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'd':
 			if (!opts.dest)
-				opts.dest = argv[optind-1];
+				opts.dest = optarg;
 			else
 				err++;
 			break;
@@ -350,7 +480,7 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'm':
 			if (!opts.match) {
-				if (!transform (argv[optind-1], &opts.match))
+				if (!transform (optarg, &opts.match))
 					err++;
 			} else {
 				err++;
@@ -358,7 +488,21 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'o':
 			if (!opts.output) {
-				opts.output = argv[optind-1];
+				opts.output = optarg;
+			} else {
+				err++;
+			}
+			break;
+		case 'O':
+		        if (!opts.optimistic) {
+				opts.optimistic++;
+			} else {
+			        err++;
+			}
+			break;
+		case 'P':
+			if (!opts.parent) {
+				opts.parent++;
 			} else {
 				err++;
 			}
@@ -384,24 +528,30 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'S':
 			if ((opts.size_begin > 0) || (opts.size_end > 0) ||
-			    !utils_parse_range (argv[optind-1], &opts.size_begin,
+			    !utils_parse_range (optarg, &opts.size_begin,
 			     &opts.size_end, TRUE)) {
 			    err++;
 			}
 			break;
 		case 't':
 			if (opts.since == 0) {
-				if (!parse_time (argv[optind-1], &opts.since))
+				if (!parse_time (optarg, &opts.since))
 					err++;
 			} else {
 			    err++;
 			}
 			break;
+		case 'T':
+			opts.truncate++;
+			break;
 		case 'u':
-			if (opts.mode == MODE_NONE) {
+			if (opts.mode == MODE_NONE)
+			{
 				end = NULL;
 				opts.mode = MODE_UNDELETE;
-				opts.uinode = strtol (optarg, &end, 0);
+				/* parse inodes */
+				if (parse_inode_arg() == -1)
+					err++;
 				if (end && *end)
 					err++;
 			} else {
@@ -444,7 +594,8 @@ int parse_options (int argc, char *argv[])
 
 		switch (opts.mode) {
 		case MODE_SCAN:
-			if (opts.output || opts.dest || (opts.fillbyte != -1)) {
+			if (opts.output || opts.dest || opts.truncate ||
+					(opts.fillbyte != (char)-1)) {
 				Eprintf ("Scan can only be used with --percent, "
 					"--match, --ignore-case, --size and --time.\n");
 				err++;
@@ -458,14 +609,16 @@ int parse_options (int argc, char *argv[])
 			if ((opts.percent != -1) || opts.match || opts.match_case ||
 			    (opts.size_begin > 0) || (opts.size_end > 0)) {
 				Eprintf ("Undelete can only be used with "
-					"--output, --destination and --byte.\n");
+					"--output, --destination, --byte and --truncate.\n");
 				err++;
 			}
 			break;
 		case MODE_COPY:
-			if ((opts.fillbyte != -1) || (opts.percent != -1) ||
+			if ((opts.fillbyte != (char)-1) || opts.truncate ||
+			    (opts.percent != -1) ||
 			    opts.match || opts.match_case ||
-			    (opts.size_begin > 0) || (opts.size_end > 0)) {
+			    (opts.size_begin > 0) ||
+			    (opts.size_end > 0)) {
 				Eprintf ("Copy can only be used with --output and --destination.\n");
 				err++;
 			}
@@ -489,9 +642,14 @@ int parse_options (int argc, char *argv[])
 				err++;
 			}
 		}
+
+		if (opts.parent && !opts.verbose) {
+			Eprintf ("To use --parent, you must also use --verbose.\n");
+			err++;
+		}
 	}
 
-	if (opts.fillbyte == -1)
+	if (opts.fillbyte == (char)-1)
 		opts.fillbyte = 0;
 
 	if (ver)
@@ -512,7 +670,7 @@ int parse_options (int argc, char *argv[])
  *
  * Return:  none
  */
-void free_file (struct ufile *file)
+static void free_file (struct ufile *file)
 {
 	struct list_head *item, *tmp;
 
@@ -524,6 +682,10 @@ void free_file (struct ufile *file)
 		Dprintf ("freeing filename '%s'\n", f->name ? f->name : NONE);
 		if (f->name)
 			free (f->name);
+		if (f->parent_name) {
+			Dprintf ("\tand parent filename '%s'\n", f->parent_name ? f->parent_name : NONE);
+			free (f->parent_name);
+		}
 		free (f);
 	}
 
@@ -541,6 +703,122 @@ void free_file (struct ufile *file)
 	free (file);
 }
 
+/**
+ * verify_parent - confirm a record is parent of a file
+ * @name:	a filename of the file
+ * @rec:	the mft record of the possible parent
+ *
+ * Check that @rec is the parent of the file represented by @name.
+ * If @rec is a directory, but it is created after @name, then we
+ * can't determine wheter @rec is really @name's parent.
+ *
+ * Return:	@rec's filename, either same name space as @name or lowest space.
+ *		NULL if can't determine parenthood or on error.
+ */
+static FILE_NAME_ATTR* verify_parent(struct filename* name, MFT_RECORD* rec) {
+	ATTR_RECORD *attr30;
+	FILE_NAME_ATTR *filename_attr = NULL, *lowest_space_name = NULL;
+	ntfs_attr_search_ctx *ctx;
+	int found_same_space = 1;
+
+	if (!name || !rec)
+		return NULL;
+
+	if (!(rec->flags & MFT_RECORD_IS_DIRECTORY)) {
+		return NULL;
+	}
+
+	ctx = ntfs_attr_get_search_ctx(NULL, rec);
+	if (!ctx) {
+		Eprintf ("Couldn't create a search context.\n");
+		return NULL;
+	}
+
+	attr30 = find_attribute(AT_FILE_NAME, ctx);
+	if (!attr30) {
+		return NULL;
+	}
+
+	filename_attr = (FILE_NAME_ATTR*)((char*)attr30 + le16_to_cpu(attr30->value_offset));
+	/* if name is older than this dir -> can't determine */
+	if (ntfs2utc(filename_attr->creation_time) > name->date_c) {
+		return NULL;
+	}
+
+	if (filename_attr->file_name_type != name->name_space) {
+		found_same_space = 0;
+		lowest_space_name = filename_attr;
+
+		while (!found_same_space && (attr30 = find_attribute(AT_FILE_NAME, ctx))) {
+			filename_attr = (FILE_NAME_ATTR*)((char*)attr30 + le16_to_cpu(attr30->value_offset));
+
+			if (filename_attr->file_name_type == name->name_space) {
+				found_same_space = 1;
+			}
+			else {
+				if (filename_attr->file_name_type < lowest_space_name->file_name_type) {
+					lowest_space_name = filename_attr;
+				}
+			}
+		}
+	}
+	
+	ntfs_attr_put_search_ctx(ctx);
+
+	return (found_same_space ? filename_attr : lowest_space_name);
+}
+
+/**
+ * get_parent_name - Find the name of a file's parent.
+ * @name:	the filename whose parent's name to find
+ *
+ */
+static void get_parent_name(struct filename* name, ntfs_volume* vol) {
+	ntfs_attr* mft_data;
+	MFT_RECORD* rec;
+	FILE_NAME_ATTR* filename_attr;
+	long long inode_num;
+
+	if (!name || !vol)
+		return;
+	
+	rec = calloc(1, vol->mft_record_size);
+	if (!rec) {
+		Eprintf ("Couldn't allocate memory in get_parent_name()\n");
+		return;
+	}
+
+	mft_data = ntfs_attr_open(vol->mft_ni, AT_DATA, NULL, 0);
+	if (!mft_data) {
+		Eprintf ("Couldn't open $MFT/$DATA: %s\n", strerror (errno));
+	}
+	else {
+		inode_num = MREF(name->parent_mref);
+
+		if (ntfs_attr_pread(mft_data, vol->mft_record_size * inode_num, vol->mft_record_size, rec) < 1) {
+			Eprintf ("Couldn't read MFT Record %lld.\n", inode_num);
+		}
+		else {
+			if ((filename_attr = verify_parent(name, rec))) {
+				if (ntfs_ucstombs(filename_attr->file_name, filename_attr->file_name_length, &name->parent_name, 0) < 0) {
+					Dprintf ("Couldn't translate filename to current locale.\n");
+					name->parent_name = NULL;
+				}
+			}
+		}
+	}
+
+	if (mft_data) {
+		ntfs_attr_close(mft_data);
+	}
+
+	if (rec) {
+		free(rec);
+	}
+
+	return;
+}
+
 /**
  * get_filenames - Read an MFT Record's $FILENAME attributes
  * @file:  The file object to work with
@@ -560,7 +838,7 @@ void free_file (struct ufile *file)
  * Return:  n  The number of $FILENAME attributes found
  *	   -1  Error
  */
-int get_filenames (struct ufile *file)
+static int get_filenames (struct ufile *file, ntfs_volume* vol)
 {
 	ATTR_RECORD *rec;
 	FILE_NAME_ATTR *attr;
@@ -604,8 +882,16 @@ int get_filenames (struct ufile *file)
 			Dprintf ("Couldn't translate filename to current locale.\n");
 		}
 
+		name->parent_name = NULL;
+
+		if (opts.parent) {
+			name->parent_mref = attr->parent_directory;
+			get_parent_name(name, vol);
+		}
+
 		if (name->name_space < space) {
 			file->pref_name = name->name;
+			file->pref_pname = name->parent_name;
 			space = name->name_space;
 		}
 
@@ -636,7 +922,7 @@ int get_filenames (struct ufile *file)
  * Return:  n  The number of $FILENAME attributes found
  *	   -1  Error
  */
-int get_data (struct ufile *file, ntfs_volume *vol)
+static int get_data (struct ufile *file, ntfs_volume *vol)
 {
 	ATTR_RECORD *rec;
 	ntfs_attr_search_ctx *ctx;
@@ -663,7 +949,7 @@ int get_data (struct ufile *file, ntfs_volume *vol)
 		data->encrypted  = rec->flags & ATTR_IS_ENCRYPTED;
 
 		if (rec->name_length) {
-			data->uname     = (uchar_t *) ((char *) rec + le16_to_cpu (rec->name_offset));
+			data->uname     = (ntfschar *) ((char *) rec + le16_to_cpu (rec->name_offset));
 			data->uname_len = rec->name_length;
 
 			if (ntfs_ucstombs (data->uname, data->uname_len, &data->name,
@@ -710,7 +996,7 @@ int get_data (struct ufile *file, ntfs_volume *vol)
  * Return:  Pointer  A ufile object containing the results
  *	    NULL     Error
  */
-struct ufile * read_record (ntfs_volume *vol, long long record)
+static struct ufile * read_record (ntfs_volume *vol, long long record)
 {
 	ATTR_RECORD *attr10, *attr20, *attr90;
 	struct ufile *file;
@@ -771,7 +1057,7 @@ struct ufile * read_record (ntfs_volume *vol, long long record)
 	if (attr90)
 		file->directory = 1;
 
-	if (get_filenames (file) < 0) {
+	if (get_filenames (file, vol) < 0) {
 		Eprintf ("Couldn't get filenames.\n");
 	}
 	if (get_data (file, vol) < 0) {
@@ -804,14 +1090,14 @@ struct ufile * read_record (ntfs_volume *vol, long long record)
  * Return:  n  The percentage of the file that _could_ be recovered
  *	   -1  Error
  */
-int calc_percentage (struct ufile *file, ntfs_volume *vol)
+static int calc_percentage (struct ufile *file, ntfs_volume *vol)
 {
 	runlist_element *rl = NULL;
 	struct list_head *pos;
 	struct data *data;
 	long long i, j;
 	long long start, end;
-	int inuse, free;
+	int clusters_inuse, clusters_free;
 	int percent = 0;
 
 	if (!file || !vol)
@@ -829,8 +1115,8 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
 
 	list_for_each (pos, &file->data) {
 		data  = list_entry (pos, struct data, list);
-		inuse = 0;
-		free  = 0;
+		clusters_inuse = 0;
+		clusters_free  = 0;
 
 		if (data->encrypted) {
 			Vprintf ("File is encrypted, recovery is impossible.\n");
@@ -863,7 +1149,7 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
 		if (rl[0].lcn == LCN_RL_NOT_MAPPED) {	/* extended mft record */
 			Vprintf ("Missing segment at beginning, %lld "
 					"clusters\n", (long long)rl[0].length);
-			inuse += rl[0].length;
+			clusters_inuse += rl[0].length;
 			rl++;
 		}
 
@@ -872,12 +1158,12 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
 				Vprintf ("Missing segment at end, %lld "
 						"clusters\n",
 						(long long)rl[i].length);
-				inuse += rl[i].length;
+				clusters_inuse += rl[i].length;
 				continue;
 			}
 
 			if (rl[i].lcn == LCN_HOLE) {
-				free += rl[i].length;
+				clusters_free += rl[i].length;
 				continue;
 			}
 
@@ -886,18 +1172,19 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
 
 			for (j = start; j < end; j++) {
 				if (utils_cluster_in_use (vol, j))
-					inuse++;
+					clusters_inuse++;
 				else
-					free++;
+					clusters_free++;
 			}
 		}
 
-		if ((inuse + free) == 0) {
+		if ((clusters_inuse + clusters_free) == 0) {
 			Eprintf ("Unexpected error whilst calculating percentage for inode %lld\n", file->inode);
 			continue;
 		}
 
-		data->percent = (free * 100) / (inuse + free);
+		data->percent = (clusters_free * 100) /
+				(clusters_inuse + clusters_free);
 
 		percent = max (percent, data->percent);
 	}
@@ -921,10 +1208,10 @@ int calc_percentage (struct ufile *file, ntfs_volume *vol)
  *
  * Return:  none
  */
-void dump_record (struct ufile *file)
+static void dump_record (struct ufile *file)
 {
 	char buffer[20];
-	char *name;
+	const char *name;
 	struct list_head *item;
 	int i;
 
@@ -960,7 +1247,13 @@ void dump_record (struct ufile *file)
 		    FILE_ATTR_COMPRESSED | FILE_ATTR_ENCRYPTED))) {
 			Qprintf (NONE);
 		}
+
 		Qprintf ("\n");
+
+		if (opts.parent) {
+			Qprintf ("Parent: %s\n", f->parent_name ? f->parent_name : "<non-determined>");
+		}
+
 		Qprintf ("Size alloc: %lld\n", f->size_alloc);
 		Qprintf ("Size data: %lld\n", f->size_data);
 
@@ -1032,11 +1325,11 @@ void dump_record (struct ufile *file)
  *
  * Return:  none
  */
-void list_record (struct ufile *file)
+static void list_record (struct ufile *file)
 {
 	char buffer[20];
 	struct list_head *item;
-	char *name = NULL;
+	const char *name = NULL;
 	long long size = 0;
 	int percent = 0;
 
@@ -1089,7 +1382,7 @@ void list_record (struct ufile *file)
  * Return:  1  There is a matching filename.
  *	    0  There is no match.
  */
-int name_match (regex_t *re, struct ufile *file)
+static int name_match (regex_t *re, struct ufile *file)
 {
 	struct list_head *item;
 	int result;
@@ -1127,7 +1420,8 @@ int name_match (regex_t *re, struct ufile *file)
  * Return:  -1  Error, something went wrong
  *	     0  Success, all the data was written
  */
-unsigned int write_data (int fd, const char *buffer, unsigned int bufsize)
+static unsigned int write_data (int fd, const char *buffer,
+	unsigned int bufsize)
 {
 	ssize_t result1, result2;
 
@@ -1174,8 +1468,8 @@ unsigned int write_data (int fd, const char *buffer, unsigned int bufsize)
  *
  * Return:  n  Length of the allocated name
  */
-int create_pathname (const char *dir, const char *name, const char *stream,
-		     char *buffer, int bufsize)
+static int create_pathname (const char *dir, const char *name,
+	const char *stream, char *buffer, int bufsize)
 {
 	if (!name)
 		name = UNKNOWN;
@@ -1205,7 +1499,7 @@ int create_pathname (const char *dir, const char *name, const char *stream,
  * Return:  -1  Error, failed to create the file
  *	     n  Success, this is the file descriptor
  */
-int open_file (const char *pathname)
+static int open_file (const char *pathname)
 {
 	int flags;
 
@@ -1229,7 +1523,7 @@ int open_file (const char *pathname)
  * Return:  1  Success, set the file's date and time
  *	    0  Error, failed to change the file's date and time
  */
-int set_date (const char *pathname, time_t date)
+static int set_date (const char *pathname, time_t date)
 {
 	struct utimbuf ut;
 
@@ -1257,14 +1551,13 @@ int set_date (const char *pathname, time_t date)
  * Return:  -1  Error, something went wrong
  *	     n  Success, the number of recoverable files
  */
-int scan_disk (ntfs_volume *vol)
+static int scan_disk (ntfs_volume *vol)
 {
 	const int BUFSIZE = 8192;
 	char *buffer = NULL;
 	int results = 0;
 	ntfs_attr *attr;
 	long long size;
-	long long read;
 	long long bmpsize;
 	int i, j, k, b;
 	int percent;
@@ -1302,8 +1595,8 @@ int scan_disk (ntfs_volume *vol)
 	Qprintf ("Inode    Flags  %%age  Date            Size  Filename\n");
 	Qprintf ("---------------------------------------------------------------\n");
 	for (i = 0; i < bmpsize; i += BUFSIZE) {
-		read = min ((bmpsize - i), BUFSIZE);
-		size = ntfs_attr_pread (attr, i, read, buffer);
+		long long read_count = min ((bmpsize - i), BUFSIZE);
+		size = ntfs_attr_pread (attr, i, read_count, buffer);
 		if (size < 0)
 			break;
 
@@ -1380,7 +1673,7 @@ out:
  * Return:  0  Error, something went wrong
  *	    1  Success, the data was recovered
  */
-int undelete_file (ntfs_volume *vol, long long inode)
+static int undelete_file (ntfs_volume *vol, long long inode)
 {
 	char pathname[256];
 	char *buffer = NULL;
@@ -1394,6 +1687,7 @@ int undelete_file (ntfs_volume *vol, long long inode)
 	long long k;
 	int result = 0;
 	char *name;
+	long long cluster_count;	/* I'll need this variable (see below). +mabs */
 
 	if (!vol)
 		return 0;
@@ -1409,6 +1703,21 @@ int undelete_file (ntfs_volume *vol, long long inode)
 	if (!buffer)
 		goto free;
 
+	if (file->mft->flags & MFT_RECORD_IN_USE) {		/* These two statement blocks were         */
+		Eprintf ("Record is in use by the mft\n");	/* relocated from below because            */
+		if (!opts.force) {				/* calc_percentage() must be called        */
+			free(buffer);
+			free_file (file);			/* before dump_record() or list_record().  */
+			return 0;				/* Otherwise, when undeleting, a file      */
+		}						/* will always be listed as 0% recoverable */
+		Vprintf ("Forced to continue.\n");		/* even if successfully undeleted.   +mabs */
+	}
+
+	if (calc_percentage (file, vol) == 0) {
+		Qprintf ("File has no recoverable data.\n");
+		goto free;
+	}
+
 	if (opts.verbose) {
 		dump_record (file);
 	} else {
@@ -1418,20 +1727,6 @@ int undelete_file (ntfs_volume *vol, long long inode)
 		Qprintf ("\n");
 	}
 
-	if (file->mft->flags & MFT_RECORD_IN_USE) {
-		Eprintf ("Record is in use by the mft\n");
-		if (!opts.force) {
-			free_file (file);
-			return 0;
-		}
-		Vprintf ("Forced to continue.\n");
-	}
-
-	if (calc_percentage (file, vol) == 0) {
-		Qprintf ("File has no recoverable data.\n");
-		goto free;
-	}
-
 	if (list_empty (&file->data)) {
 		Qprintf ("File has no data.  There is nothing to recover.\n");
 		goto free;
@@ -1496,6 +1791,7 @@ int undelete_file (ntfs_volume *vol, long long inode)
 				}
 			}
 
+			cluster_count = 0LL;
 			for (i = 0; rl[i].length > 0; i++) {
 
 				if (rl[i].lcn == LCN_RL_NOT_MAPPED) {
@@ -1509,6 +1805,7 @@ int undelete_file (ntfs_volume *vol, long long inode)
 							close (fd);
 							goto free;
 						}
+						cluster_count++;
 					}
 					continue;
 				}
@@ -1530,7 +1827,7 @@ int undelete_file (ntfs_volume *vol, long long inode)
 				end   = rl[i].lcn + rl[i].length;
 
 				for (j = start; j < end; j++) {
-					if (utils_cluster_in_use (vol, j)) {
+					if (utils_cluster_in_use (vol, j) && !opts.optimistic) {
 						memset (buffer, opts.fillbyte, bufsize);
 						if (write_data (fd, buffer, bufsize) < bufsize) {
 							Eprintf ("Write failed: %s\n", strerror (errno));
@@ -1548,10 +1845,35 @@ int undelete_file (ntfs_volume *vol, long long inode)
 							close (fd);
 							goto free;
 						}
+						cluster_count++;
 					}
 				}
 			}
 			Qprintf ("\n");
+
+	/* The following block of code implements the --truncate option.        */
+	/* Its semantics are as follows:                                        */
+	/* IF opts.truncate is set AND data stream currently being recovered is */
+	/* non-resident AND data stream has no holes (100% recoverability) AND  */
+	/* 0 <= (data->size_alloc - data->size_data) <= vol->cluster_size AND   */
+	/* cluster_count * vol->cluster_size == data->size_alloc THEN file      */
+	/* currently being written is truncated to data->size_data bytes before */
+	/* it's closed.                                                         */
+	/* This multiple checks try to ensure that only files with consistent   */
+	/* values of size/occupied clusters are eligible for truncation. Note   */
+	/* that resident streams need not be truncated, since the original code */
+	/* already recovers their exact length.                           +mabs */
+
+			if (opts.truncate) {
+				if (d->percent == 100 && d->size_alloc >= d->size_data &&
+					(d->size_alloc - d->size_data) <= (long long)vol->cluster_size &&
+					cluster_count * (long long)vol->cluster_size == d->size_alloc) {
+					if (ftruncate(fd, (off_t)d->size_data))
+						Eprintf("Truncation failed: %s\n", strerror(errno));
+				} else Qprintf("Truncation not performed because file has an "
+					"inconsistent $MFT record.\n");
+			}
+
 			if (close (fd) < 0) {
 				Eprintf ("Close failed: %s\n", strerror (errno));
 			}
@@ -1583,7 +1905,7 @@ free:
  * Return:  0  Success, all the records were written
  *	    1  Error, something went wrong
  */
-int copy_mft (ntfs_volume *vol, long long mft_begin, long long mft_end)
+static int copy_mft (ntfs_volume *vol, long long mft_begin, long long mft_end)
 {
 	char pathname[256];
 	ntfs_attr *mft;
@@ -1668,6 +1990,8 @@ int main (int argc, char *argv[])
 {
 	ntfs_volume *vol;
 	int result = 1;
+	int i;
+	u32 inode;
 
 	if (!parse_options (argc, argv))
 		goto free;
@@ -1685,17 +2009,29 @@ int main (int argc, char *argv[])
 			Vprintf ("Failed to scan device '%s'.\n", opts.device);
 		break;
 	case MODE_UNDELETE:
-		result = !undelete_file (vol, opts.uinode);
-		if (result)
-			Vprintf ("Failed to undelete inode %d.\n", opts.uinode);
+		if (nr_entries == 0)
+		{
+			printf ("ERROR: No inode(s) specified!\n");
+			break;
+		}
+
+		for (i = 0; i < nr_entries; i++)
+		{
+			for (inode = ranges[i].begin; inode <= ranges[i].end; inode ++)
+			{
+				result = !undelete_file (vol, inode);
+				if (result)
+					Vprintf ("ERROR: Failed to undelete "
+							"inode %u\n!",
+							(unsigned int)inode);
+			}
+		}
 		break;
 	case MODE_COPY:
 		result = !copy_mft (vol, opts.mft_begin, opts.mft_end);
 		if (result)
 			Vprintf ("Failed to read MFT blocks %lld-%lld.\n",
-					(long long)opts.mft_begin,
-					(long long)min(vol->nr_mft_records,
-					opts.mft_end));
+				opts.mft_begin, min(vol->nr_mft_records, opts.mft_end));
 		break;
 	default:
 		; /* Cannot happen */
