@@ -2,6 +2,7 @@
  * ntfswipe - Part of the Linux-NTFS project.
  *
  * Copyright (c) 2002-2003 Richard Russon
+ * Copyright (c) 2004 Yura Pakhuchiy
  *
  * This utility will overwrite unused space on an NTFS volume.
  *
@@ -36,8 +37,10 @@
 #include "volume.h"
 #include "utils.h"
 #include "debug.h"
+#include "dir.h"
+#include "mst.h"
 
-static char *EXEC_NAME = "ntfswipe";
+static const char *EXEC_NAME = "ntfswipe";
 static struct options opts;
 
 GEN_PRINTF (Eprintf, stderr, NULL,          FALSE)
@@ -51,11 +54,12 @@ GEN_PRINTF (Qprintf, stdout, &opts.quiet,   FALSE)
  *
  * Return:  none
  */
-void version (void)
+static void version (void)
 {
 	printf ("\n%s v%s - Overwrite the unused space on an NTFS Volume.\n\n",
 		EXEC_NAME, VERSION);
 	printf ("Copyright (c) 2002-2003 Richard Russon\n");
+	printf ("Copyright (c) 2004 Yura Pakhuchiy\n");
 	printf ("\n%s\n%s%s\n", ntfs_gpl, ntfs_bugs, ntfs_home);
 }
 
@@ -66,7 +70,7 @@ void version (void)
  *
  * Return:  none
  */
-void usage (void)
+static void usage (void)
 {
 	printf ("\nUsage: %s [options] device\n"
 		"    -i       --info        Show volume information (default)\n"
@@ -107,7 +111,7 @@ void usage (void)
  * Return:  0  Error, invalid string
  *	    n  Success, the count of numbers parsed
  */
-int parse_list (char *list, int **result)
+static int parse_list (char *list, int **result)
 {
 	char *ptr;
 	char *end;
@@ -167,9 +171,9 @@ int parse_list (char *list, int **result)
  * Return:  1 Success
  *	    0 Error, one or more problems
  */
-int parse_options (int argc, char *argv[])
+static int parse_options (int argc, char *argv[])
 {
-	static char *sopt = "-ab:c:dfh?ilmnpqtuvV";
+	static const char *sopt = "-ab:c:dfh?ilmnpqtuvV";
 	static struct option lopt[] = {
 		{ "all",	no_argument,		NULL, 'a' },
 		{ "bytes",	required_argument,	NULL, 'b' },
@@ -224,7 +228,7 @@ int parse_options (int argc, char *argv[])
 			break;
 		case 'b':
 			if (!opts.bytes) {
-				if (!parse_list (argv[optind-1], &opts.bytes))
+				if (!parse_list (optarg, &opts.bytes))
 					err++;
 			} else {
 				err++;
@@ -344,13 +348,15 @@ int parse_options (int argc, char *argv[])
  * wipe_unused - Wipe unused clusters
  * @vol:   An ntfs volume obtained from ntfs_mount
  * @byte:  Overwrite with this value
+ * @act:   Wipe, test or info
  *
  * Read $Bitmap and wipe any clusters that are marked as not in use.
  *
- * Return:  1  Success, the clusters were wiped
- *	    0  Error, something went wrong
+ * Return: >0  Success, the atrribute was wiped
+ *          0  Nothing to wipe
+ *         -1  Error, something went wrong
  */
-s64 wipe_unused (ntfs_volume *vol, int byte, enum action act)
+static s64 wipe_unused (ntfs_volume *vol, int byte, enum action act)
 {
 	s64 i;
 	s64 total = 0;
@@ -393,38 +399,250 @@ free:
 	return total;
 }
 
+/**
+ * wipe_compressed_attribute - Wipe compressed $DATA attribute
+ * @vol:	An ntfs volume obtained from ntfs_mount
+ * @byte:	Overwrite with this value
+ * @act:	Wipe, test or info
+ * @na:		Opened ntfs attribute
+ *
+ * Return: >0  Success, the atrribute was wiped
+ *          0  Nothing to wipe
+ *         -1  Error, something went wrong
+ */
+static s64 wipe_compressed_attribute (ntfs_volume *vol, int byte,
+						enum action act, ntfs_attr *na)
+{
+	unsigned char *buf;
+	s64 size, offset, ret, wiped = 0;
+	u16 block_size;
+	VCN cur_vcn = 0;
+	runlist *rlc = na->rl;
+	s64 cu_mask = na->compression_block_clusters - 1;
+		
+	while (rlc->length) {
+		cur_vcn += rlc->length;
+		if ((cur_vcn & cu_mask) ||
+			(((rlc + 1)->length) && (rlc->lcn != LCN_HOLE))) {
+			rlc++;
+			continue;
+		}
+
+		if (rlc->lcn == LCN_HOLE) {
+			runlist *rlt;
+
+			offset = cur_vcn - rlc->length;
+			if (offset == (offset & (~cu_mask))) {
+				rlc++;
+				continue;
+			}
+			offset = (offset & (~cu_mask))
+						<< vol->cluster_size_bits;
+			rlt = rlc;
+			while ((rlt - 1)->lcn == LCN_HOLE) rlt--;
+			while (1) {
+				ret = ntfs_rl_pread (vol, na->rl,
+						offset, 2, &block_size);
+				block_size = le16_to_cpu (block_size);
+				if (ret != 2) {
+					Vprintf ("Internal error\n");
+					Eprintf ("ntfs_rl_pread failed");
+					return -1;
+				}
+				if (block_size == 0) {
+					offset += 2;
+					break;
+				}
+				block_size &= 0x0FFF;
+				block_size += 3;
+				offset += block_size;
+				if (offset >= (((rlt->vcn) <<
+						vol->cluster_size_bits) - 2))
+					goto next;
+			}
+			size = (rlt->vcn << vol->cluster_size_bits) - offset;
+		} else {
+			size = na->allocated_size - na->data_size;
+			offset = (cur_vcn << vol->cluster_size_bits) - size;
+		}
+		
+		if (size < 0) {
+			Vprintf ("Internal error\n");
+			Eprintf ("bug or damaged fs: we want "
+				"allocate buffer size %lld bytes", size);
+			return -1;
+		}
+		
+		if ((act == act_info) || (!size)) {
+			wiped += size;
+			rlc++;
+			continue;
+		}
+		
+		buf = malloc (size);
+		if (!buf) {
+			Vprintf ("Not enough memory\n");
+			Eprintf ("Not enough memory to allocate "
+							"%lld bytes", size);
+			return -1;
+		}
+		memset (buf, byte, size);
+		
+		ret = ntfs_rl_pwrite (vol, na->rl, offset, size, buf);
+		free (buf);
+		if (ret != size) {
+			Vprintf ("Internal error\n");
+			Eprintf ("ntfs_rl_pwrite failed, offset %llu, "
+				"size %lld, vcn %lld",	offset, size, rlc->vcn);
+			return -1;
+		}
+		wiped += ret;
+next:
+		rlc++;
+	}
+	
+	return wiped;
+}
+
+/**
+ * wipe_attribute - Wipe not compressed $DATA attribute
+ * @vol:	An ntfs volume obtained from ntfs_mount
+ * @byte:	Overwrite with this value
+ * @act:	Wipe, test or info
+ * @na:		Opened ntfs attribute
+ *
+ * Return: >0  Success, the atrribute was wiped
+ *          0  Nothing to wipe
+ *         -1  Error, something went wrong
+ */
+static s64 wipe_attribute (ntfs_volume *vol, int byte, enum action act, 
+								ntfs_attr *na)
+{
+	unsigned char *buf;
+	s64 wiped;
+	s64 size;
+	u64 offset = na->data_size;
+
+	if (!offset)
+		return 0;
+	if (NAttrEncrypted(na))
+		offset = (((offset - 1) >> 10) + 1) << 10;
+	size = (vol->cluster_size - offset) % vol->cluster_size;
+	
+	if (act == act_info)
+		return size;
+
+	buf = malloc (size);
+	if (!buf) {
+		Vprintf ("Not enough memory\n");
+		Eprintf ("Not enough memory to allocate %lld bytes", size);
+		return -1;
+	}
+	memset (buf, byte, size);
+
+	wiped = ntfs_rl_pwrite (vol, na->rl, offset, size, buf);
+	if (wiped == -1) {
+		Vprintf ("Internal error\n");
+		Eprintf ("Couldn't wipe tail");
+	}
+
+	free (buf);
+	return wiped;
+}
+
 /**
  * wipe_tails - Wipe the file tails
  * @vol:   An ntfs volume obtained from ntfs_mount
  * @byte:  Overwrite with this value
+ * @act:   Wipe, test or info
  *
  * Disk space is allocated in clusters.  If a file isn't an exact multiple of
  * the cluster size, there is some slack space at the end.  Wipe this space.
  *
- * Return:  1  Success, the clusters were wiped
- *	    0  Error, something went wrong
+ * Return: >0  Success, the clusters were wiped
+ *          0  Nothing to wipe
+ *         -1  Error, something went wrong
  */
-s64 wipe_tails (ntfs_volume *vol, int byte, enum action act)
+static s64 wipe_tails (ntfs_volume *vol, int byte, enum action act)
 {
+	s64 total = 0;
+	s64 inode_num;
+	ntfs_inode *ni;
+	ntfs_attr *na;
+
 	if (!vol || (byte < 0))
 		return -1;
 
-	Qprintf ("wipe_tails (not implemented) 0x%02x\n", byte);
-	return 0;
+	for (inode_num = 16; inode_num < vol->nr_mft_records; inode_num++) {
+		s64 wiped;
+
+		Vprintf ("Inode %lld - ", inode_num);
+		ni = ntfs_inode_open (vol, inode_num);
+		if (!ni) {
+			Vprintf ("Could not open inode\n");
+			continue;
+		}
+
+		if (ni->mrec->base_mft_record) {
+			Vprintf ("Not base mft record. Skipping\n");
+			goto close_inode;
+		}
+
+		na = ntfs_attr_open (ni, AT_DATA, AT_UNNAMED, 0);
+		if (!na) {
+			Vprintf ("Couldn't open $DATA attribute\n");
+			goto close_inode;
+		}
+
+		if (!NAttrNonResident(na)) {
+			Vprintf ("Resident $DATA atrribute. Skipping.\n");
+			goto close_attr;
+		}
+
+		if (ntfs_attr_map_whole_runlist(na)) {
+			Vprintf ("Internal error\n");
+			Eprintf ("Can't map runlist (inode %lld)\n", inode_num);
+			goto close_attr;
+		}
+
+		if (NAttrCompressed(na))
+			wiped = wipe_compressed_attribute (vol, byte, act, na);
+		else
+			wiped = wipe_attribute (vol, byte, act, na);
+
+		if (wiped == -1) {
+			Eprintf (" (inode %lld)\n", inode_num);
+			goto close_attr;
+		}
+		
+		if (wiped) {
+			Vprintf ("Wiped %llu bytes\n", wiped);
+			total += wiped;
+		} else
+			Vprintf ("Nothing to wipe\n");
+close_attr:
+		ntfs_attr_close (na);
+close_inode:
+		ntfs_inode_close (ni);
+	}
+	Qprintf ("wipe_tails 0x%02x, %lld bytes\n", byte, total);
+	return total;
 }
 
 /**
  * wipe_mft - Wipe the MFT slack space
  * @vol:   An ntfs volume obtained from ntfs_mount
  * @byte:  Overwrite with this value
+ * @act:   Wipe, test or info
  *
  * MFT Records are 1024 bytes long, but some of this space isn't used.  Wipe any
  * unused space at the end of the record and wipe any unused records.
  *
- * Return:  1  Success, the clusters were wiped
- *	    0  Error, something went wrong
+ * Return: >0  Success, the clusters were wiped
+ *          0  Nothing to wipe
+ *         -1  Error, something went wrong
  */
-s64 wipe_mft (ntfs_volume *vol, int byte, enum action act)
+static s64 wipe_mft (ntfs_volume *vol, int byte, enum action act)
 {
 	// by considering the individual attributes we might be able to
 	// wipe a few more bytes at the attr's tail.
@@ -543,38 +761,290 @@ free:
 	return total;
 }
 
+/**
+ * wipe_index_allocation - Wipe $INDEX_ALLOCATION attribute
+ * @vol:		An ntfs volume obtained from ntfs_mount
+ * @byte:		Overwrite with this value
+ * @act:		Wipe, test or info
+ * @naa:		Opened ntfs $INDEX_ALLOCATION attribute
+ * @nab:		Opened ntfs $BIMTAP attribute
+ * @indx_record_size:	Size of INDX record
+ *
+ * Return: >0  Success, the clusters were wiped
+ *          0  Nothing to wipe
+ *         -1  Error, something went wrong
+ */
+static s64 wipe_index_allocation (ntfs_volume *vol, int byte, enum action act,
+			ntfs_attr *naa, ntfs_attr *nab, u32 indx_record_size) {
+	s64 total = 0;
+	s64 wiped = 0;
+	s64 offset = 0;
+	s64 obyte = 0;
+	u64 wipe_offset;
+	s64 wipe_size;
+	u8 obit = 0;
+	u8 mask;
+	u8 *bitmap;
+	u8 *buf;
+	
+	bitmap = malloc (nab->data_size);
+	if (!bitmap) {
+		Vprintf ("malloc failed\n");
+		Eprintf ("Couldn't allocate %lld bytes", nab->data_size);
+		return -1;
+	}
+	
+	if (ntfs_attr_pread (nab, 0, nab->data_size, bitmap)
+						!= nab->data_size) {
+		Vprintf ("Internal error\n");
+		Eprintf ("Couldn't read $BITMAP");
+		total = -1;
+		goto free_bitmap;
+	}
+	
+	buf = malloc (indx_record_size);
+	if (!buf) {
+		Vprintf ("malloc failed\n");
+		Eprintf ("Couldn't allocate %u bytes",
+				(unsigned int)indx_record_size);
+		total = -1;
+		goto free_bitmap;
+	}
+	
+	while (offset < naa->allocated_size) {
+		mask = 1 << obit;
+		if (bitmap[obyte] & mask) {
+			INDEX_ALLOCATION *indx;
+			
+			s64 ret = ntfs_rl_pread (vol, naa->rl,
+					offset, indx_record_size, buf);
+			if (ret != indx_record_size) {
+				Vprintf ("ntfs_rl_pread failed\n");
+				Eprintf ("Couldn't read INDX record");
+				total = -1;
+				goto free_buf;
+			}
+			
+			indx = (INDEX_ALLOCATION *) buf;
+			if (ntfs_mst_post_read_fixup ((NTFS_RECORD *)buf,
+								indx_record_size))
+				Eprintf ("damaged fs: mst_post_read_fixup failed");
+    			
+			if ((le32_to_cpu(indx->index.allocated_size) + 0x18) != 
+							indx_record_size) {
+				Vprintf ("Internal error\n");
+				Eprintf ("INDX record should be %u bytes",
+						(unsigned int)indx_record_size);
+				total = -1;
+				goto free_buf;
+			}
+			
+			wipe_offset = le32_to_cpu(indx->index.index_length) + 0x18;
+			wipe_size = indx_record_size - wipe_offset;
+			memset (buf + wipe_offset, byte, wipe_size);
+			if (ntfs_mst_pre_write_fixup ((NTFS_RECORD *)indx,
+								indx_record_size))
+				Eprintf ("damaged fs: mst_pre_write_protect failed");
+			if (opts.verbose > 1)
+				Vprintf ("+");
+		} else {
+			wipe_size = indx_record_size;
+			memset (buf, byte, wipe_size);
+			if (opts.verbose > 1)
+				Vprintf ("x");
+		}
+		
+		wiped = ntfs_rl_pwrite (vol, naa->rl, offset, indx_record_size, buf);
+		if (wiped != indx_record_size) {
+			Vprintf ("ntfs_rl_pwrite failed\n");
+			Eprintf ("Couldn't wipe tail of INDX record");
+			total = -1;
+			goto free_buf;
+		}
+		total += wipe_size;
+
+		offset += indx_record_size;
+		obit++;
+		if (obit > 7) {
+			obit = 0;
+			obyte++;
+		}
+	}
+	if ((opts.verbose > 1) && (wiped != -1))
+		Vprintf ("\n\t");
+free_buf:
+	free (buf);
+free_bitmap:
+	free (bitmap);
+	return total;
+}
+
+/**
+ * get_indx_record_size - determine size of INDX record from $INDEX_ROOT
+ * @nar:	Opened ntfs $INDEX_ROOT attribute
+ *
+ * Return: >0  Success, return INDX record size
+ *          0  Error, something went wrong
+ */
+static u32 get_indx_record_size (ntfs_attr *nar)
+{
+	u32 indx_record_size;
+	
+	if (ntfs_attr_pread (nar, 8, 4, &indx_record_size) != 4) {
+		Vprintf ("Couldn't determine size of INDX record\n");
+		Eprintf ("ntfs_attr_pread failed");
+		return 0;
+	}
+	
+	indx_record_size = le32_to_cpu (indx_record_size);
+	if (!indx_record_size) {
+		Vprintf ("Internal error\n");
+		Eprintf ("INDX record should be 0");
+	}
+	return indx_record_size;
+}
+
 /**
  * wipe_directory - Wipe the directory indexes
- * @vol:   An ntfs volume obtained from ntfs_mount
- * @byte:  Overwrite with this value
+ * @vol:	An ntfs volume obtained from ntfs_mount
+ * @byte:	Overwrite with this value
+ * @act:	Wipe, test or info
  *
  * Directories are kept in sorted B+ Trees.  Index blocks may not be full.  Wipe
  * the unused space at the ends of these blocks.
  *
- * Return:  1  Success, the clusters were wiped
- *	    0  Error, something went wrong
+ * Return: >0  Success, the clusters were wiped
+ *          0  Nothing to wipe
+ *         -1  Error, something went wrong
  */
-s64 wipe_directory (ntfs_volume *vol, int byte, enum action act)
+static s64 wipe_directory (ntfs_volume *vol, int byte, enum action act)
 {
+	s64 total = 0;
+	s64 inode_num;
+	ntfs_inode *ni;
+	ntfs_attr *naa;
+	ntfs_attr *nab;
+	ntfs_attr *nar;
+
 	if (!vol || (byte < 0))
 		return -1;
 
-	Qprintf ("wipe_directory (not implemented) 0x%02x\n", byte);
-	return 0;
+	for (inode_num = 5; inode_num < vol->nr_mft_records; inode_num++) {
+		u32 indx_record_size;
+		s64 wiped;
+
+		Vprintf ("Inode %lld - ", inode_num);
+		ni = ntfs_inode_open (vol, inode_num);
+		if (!ni) {
+			if (opts.verbose > 2)
+				Vprintf ("Could not open inode\n");
+			else
+				Vprintf ("\r");
+			continue;
+		}
+
+		if (ni->mrec->base_mft_record) {
+			if (opts.verbose > 2)
+				Vprintf ("Not base mft record. Skipping\n");
+			else
+				Vprintf ("\r");
+			goto close_inode;
+		}
+
+		naa = ntfs_attr_open (ni, AT_INDEX_ALLOCATION, I30, 4);
+		if (!naa) {
+			if (opts.verbose > 2)
+				Vprintf ("Couldn't open $INDEX_ALLOCATION\n");
+			else
+				Vprintf ("\r");
+			goto close_inode;
+		}
+		
+		if (!NAttrNonResident(naa)) {
+			Vprintf ("Resident $INDEX_ALLOCATION\n");
+			Eprintf ("damaged fs: Resident $INDEX_ALLOCATION "
+					"(inode %lld)\n", inode_num);
+			goto close_attr_allocation;
+		}
+
+		if (ntfs_attr_map_whole_runlist(naa)) {
+			Vprintf ("Internal error\n");
+			Eprintf ("Can't map runlist for $INDEX_ALLOCATION "
+					"(inode %lld)\n", inode_num);
+			goto close_attr_allocation;
+		}
+
+		nab = ntfs_attr_open (ni, AT_BITMAP, I30, 4);
+		if (!nab) {
+			Vprintf ("Couldn't open $BITMAP\n");
+			Eprintf ("damaged fs: $INDEX_ALLOCATION is present, "
+					"but we can't open $BITMAP with same "
+					"name (inode %lld)\n", inode_num);
+			goto close_attr_allocation;
+		}
+		
+		nar = ntfs_attr_open (ni, AT_INDEX_ROOT, I30, 4);
+		if (!nar) {
+			Vprintf ("Couldn't open $INDEX_ROOT\n");
+			Eprintf ("damaged fs: $INDEX_ALLOCATION is present, but "
+					"we can't open $INDEX_ROOT with same name"
+					" (inode %lld)\n", inode_num);
+			goto close_attr_bitmap;
+		}
+		
+		if (NAttrNonResident(nar)) {
+			Vprintf ("Not resident $INDEX_ROOT\n");
+			Eprintf ("damaged fs: Not resident $INDEX_ROOT "
+					"(inode %lld)\n", inode_num);
+			goto close_attr_root;
+		}
+		
+		indx_record_size = get_indx_record_size (nar);
+		if (!indx_record_size) {
+			Eprintf (" (inode %lld)\n", inode_num);
+			goto close_attr_root;
+		}
+		
+		wiped = wipe_index_allocation (vol, byte, act,
+						naa, nab, indx_record_size);
+		if (wiped == -1) {
+			Eprintf (" (inode %lld)\n", inode_num);
+			goto close_attr_root;
+		}
+		
+		if (wiped) {
+			Vprintf ("Wiped %llu bytes\n", wiped);
+			total += wiped;
+		} else
+			Vprintf ("Nothing to wipe\n");
+close_attr_root:
+		ntfs_attr_close (nar);
+close_attr_bitmap:
+		ntfs_attr_close (nab);
+close_attr_allocation:
+		ntfs_attr_close (naa);
+close_inode:
+		ntfs_inode_close (ni);
+	}
+
+	Qprintf ("wipe_directory 0x%02x, %lld bytes\n", byte, total);
+	return total;
 }
 
 /**
  * wipe_logfile - Wipe the logfile (journal)
  * @vol:   An ntfs volume obtained from ntfs_mount
  * @byte:  Overwrite with this value
+ * @act:   Wipe, test or info
  *
  * The logfile journals the metadata to give the volume fault-tolerance.  If the
  * volume is in a consistant state, then this information can be erased.
  *
- * Return:  1  Success, the clusters were wiped
- *	    0  Error, something went wrong
+ * Return: >0  Success, the clusters were wiped
+ *          0  Nothing to wipe
+ *         -1  Error, something went wrong
  */
-s64 wipe_logfile (ntfs_volume *vol, int byte, enum action act)
+static s64 wipe_logfile (ntfs_volume *vol, int byte, enum action act)
 {
 	const int NTFS_BUF_SIZE2 = 8192;
 	//FIXME(?): We might need to zero the LSN field of every single mft
@@ -630,91 +1100,7 @@ s64 wipe_logfile (ntfs_volume *vol, int byte, enum action act)
 		goto io_error_exit;
 	}
 
-	/* Fill the buffer with 0xff's. */
-	memset(buf, -1, NTFS_BUF_SIZE2);
-
-	/* Set the $DATA attribute. */
-	pos = 0;
-	while ((count = len - pos) > 0) {
-		if (count > NTFS_BUF_SIZE2)
-			count = NTFS_BUF_SIZE2;
-
-		if ((count = ntfs_attr_pwrite(na, pos, count, buf)) <= 0) {
-			Dprintf("Failed to set the $LogFile attribute value.");
-			if (count != -1)
-				errno = EIO;
-			goto io_error_exit;
-		}
-
-		pos += count;
-	}
-
-	ntfs_attr_close(na);
-	return ntfs_inode_close(ni);
-
-io_error_exit:
-	eo = errno;
-	ntfs_attr_close(na);
-	errno = eo;
-error_exit:
-	eo = errno;
-	ntfs_inode_close(ni);
-	errno = eo;
-	return -1;
-	return 0;
-}
-
-/**
- * wipe_pagefile - Wipe the pagefile (swap space)
- * @vol:   An ntfs volume obtained from ntfs_mount
- * @byte:  Overwrite with this value
- *
- * pagefile.sys is used by Windows as extra virtual memory (swap space).
- * Windows recreates the file at bootup, so it can be wiped without harm.
- *
- * Return:  1  Success, the clusters were wiped
- *	    0  Error, something went wrong
- */
-s64 wipe_pagefile (ntfs_volume *vol, int byte, enum action act)
-{
-	// wipe completely, chkdsk doesn't do anything, booting writes header
-	const int NTFS_BUF_SIZE2 = 4096;
-	ntfs_inode *ni;
-	ntfs_attr *na;
-	s64 len, pos, count;
-	char buf[NTFS_BUF_SIZE2];
-	int eo;
-
-	if (!vol || (byte < 0))
-		return -1;
-
-	//Qprintf ("wipe_pagefile (not implemented) 0x%02x\n", byte);
-	
-	ni = utils_pathname_to_inode (vol, NULL, "pagefile.sys");
-	if (!ni) {
-		Dprintf("Failed to open inode FILE_LogFile.\n");
-		return -1;
-	}
-
-	if ((na = ntfs_attr_open(ni, AT_DATA, AT_UNNAMED, 0)) == NULL) {
-		Dprintf("Failed to open $FILE_LogFile/$DATA\n");
-		goto error_exit;
-	}
-
-	/* The $DATA attribute of the $LogFile has to be non-resident. */
-	if (!NAttrNonResident(na)) {
-		Dprintf("$LogFile $DATA attribute is resident!?!\n");
-		errno = EIO;
-		goto io_error_exit;
-	}
-
-	/* Get length of $LogFile contents. */
-	len = na->data_size;
-	if (!len) {
-		Dprintf("$LogFile has zero length, no disk write needed.\n");
-		return 0;
-	}
-
+	/* Fill the buffer with @byte's. */
 	memset(buf, byte, NTFS_BUF_SIZE2);
 
 	/* Set the $DATA attribute. */
@@ -734,7 +1120,9 @@ s64 wipe_pagefile (ntfs_volume *vol, int byte, enum action act)
 	}
 
 	ntfs_attr_close(na);
-	return ntfs_inode_close(ni);
+	ntfs_inode_close(ni);
+	Qprintf ("wipe_logfile 0x%02x, %lld bytes\n", byte, pos);
+	return pos;
 
 io_error_exit:
 	eo = errno;
@@ -748,181 +1136,90 @@ error_exit:
 }
 
 /**
- * ntfs_info - Display information about the NTFS Volume
- * @vol:  An ntfs volume obtained from ntfs_mount
+ * wipe_pagefile - Wipe the pagefile (swap space)
+ * @vol:   An ntfs volume obtained from ntfs_mount
+ * @byte:  Overwrite with this value
+ * @act:   Wipe, test or info
  *
- * Tell the user how much could be cleaned up.  List the number of free
- * clusters, MFT records, etc.
+ * pagefile.sys is used by Windows as extra virtual memory (swap space).
+ * Windows recreates the file at bootup, so it can be wiped without harm.
  *
- * Return:  1  Success, displayed some info
- *	    0  Error, something went wrong
+ * Return: >0  Success, the clusters were wiped
+ *          0  Nothing to wipe
+ *         -1  Error, something went wrong
  */
-int ntfs_info (ntfs_volume *vol)
+static s64 wipe_pagefile (ntfs_volume *vol, int byte, enum action act)
 {
-	u8 *buffer;
+	// wipe completely, chkdsk doesn't do anything, booting writes header
+	const int NTFS_BUF_SIZE2 = 4096;
+	ntfs_inode *ni;
+	ntfs_attr *na;
+	s64 len, pos, count;
+	char buf[NTFS_BUF_SIZE2];
+	int eo;
 
-	if (!vol)
+	if (!vol || (byte < 0))
+		return -1;
+
+	//Qprintf ("wipe_pagefile (not implemented) 0x%02x\n", byte);
+	
+	ni = utils_pathname_to_inode (vol, NULL, "pagefile.sys");
+	if (!ni) {
+		Dprintf("Failed to open inode of pagefile.sys.\n");
 		return 0;
-
-	Qprintf ("ntfs_info\n");
-
-	Qprintf ("\n");
-
-	Qprintf ("Cluster size = %u\n", vol->cluster_size);
-	Qprintf ("Volume size = %lld clusters\n", (long long)vol->nr_clusters);
-	Qprintf ("Volume size = %lld bytes\n",
-			(long long)vol->nr_clusters * vol->cluster_size);
-	Qprintf ("Volume size = %lld MiB\n", (long long)vol->nr_clusters *
-			vol->cluster_size / (1024*1024)); /* round up? */
-
-	Qprintf ("\n");
-
-	// move back bufsize
-	buffer = malloc (vol->mft_record_size);
-	if (!buffer)
-		return 0;
-
-	Qprintf ("cluster\n");
-	//Qprintf ("allocated_size = %lld\n", vol->lcnbmp_na->allocated_size);
-	Qprintf ("data_size = %lld\n", (long long)vol->lcnbmp_na->data_size);
-	//Qprintf ("initialized_size = %lld\n", vol->lcnbmp_na->initialized_size);
-
-	{
-	s64 offset;
-	s64 size = vol->lcnbmp_na->allocated_size;
-	int bufsize = vol->mft_record_size;
-	s64 use = 0;
-	s64 not = 0;
-	int i, j;
-
-	for (offset = 0; offset < size; offset += bufsize) {
-
-		if ((offset + bufsize) > size)
-			bufsize = size - offset;
-
-		if (ntfs_attr_pread (vol->lcnbmp_na, offset, bufsize, buffer) < bufsize) {
-			Eprintf ("error\n");
-			return 0;
-		}
-
-		for (i = 0; i < bufsize; i++) {
-			for (j = 0; j < 8; j++) {
-				if ((((offset+i)*8) + j) >= vol->nr_clusters)
-					goto done;
-				if (buffer[i] & (1 << j)) {
-					//printf ("*");
-					use++;
-				} else {
-					//printf (".");
-					not++;
-				}
-			}
-		}
-	}
-done:
-
-	Qprintf ("cluster use %lld, not %lld, total %lld\n", (long long)use,
-			(long long)not, (long long)(use + not));
-	Qprintf ("\n");
-
 	}
 
-	{
-	u8 *bitmap;
-	s64 bmpoff;
-	s64 bmpsize = vol->mftbmp_na->data_size;
-	int bmpbufsize = 512;
-	int i, j;
-	s64 use = 0, not = 0;
-
-	bitmap = malloc (bmpbufsize);
-	if (!bitmap)
-		return 0;
-
-	printf ("mft has %lld records\n", (long long)vol->nr_mft_records);
-
-	//Qprintf ("allocated_size = %lld\n", vol->mftbmp_na->allocated_size);
-	Qprintf ("data_size = %lld\n", (long long)vol->mftbmp_na->data_size);
-	//Qprintf ("initialized_size = %lld\n", vol->mftbmp_na->initialized_size);
-
-	printf ("bmpsize = %lld\n", (long long)bmpsize);
-	for (bmpoff = 0; bmpoff < bmpsize; bmpoff += bmpbufsize) {
-		if ((bmpoff + bmpbufsize) > bmpsize)
-			bmpbufsize = bmpsize - bmpoff;
-
-		//printf ("bmpbufsize = %d\n", bmpbufsize);
-
-		if (ntfs_attr_pread (vol->mftbmp_na, bmpoff, bmpbufsize, bitmap) < bmpbufsize) {
-			Eprintf ("error\n");
-			return 0;
-		}
-
-		for (i = 0; i < bmpbufsize; i++) {
-			for (j = 0; j < 8; j++) {
-				if ((((bmpoff+i)*8) + j) >= vol->nr_mft_records)
-					goto bmpdone;
-				if (bitmap[i] & (1 << j)) {
-					//printf ("*");
-					use++;
-				} else {
-					//printf (".");
-					not++;
-				}
-			}
-		}
+	if ((na = ntfs_attr_open(ni, AT_DATA, AT_UNNAMED, 0)) == NULL) {
+		Dprintf("Failed to open pagefile.sys/$DATA\n");
+		goto error_exit;
 	}
 
-bmpdone:
-	printf ("mft\n");
-	printf ("use %lld, not %lld, total %lld\n", (long long)use,
-			(long long)not, (long long)(use + not));
-
-	free (bitmap);
+	/* The $DATA attribute of the pagefile.sys has to be non-resident. */
+	if (!NAttrNonResident(na)) {
+		Dprintf("pagefile.sys $DATA attribute is resident!?!\n");
+		errno = EIO;
+		goto io_error_exit;
 	}
 
-	/*
-	 * wipe_unused - volume = n clusters, u unused (%age & MB)
-	 *	$Bitmap
-	 *	vol->lcnbmp_na
-	 *
-	 * wipe_tails - volume = n files, total tail slack space
-	 *	$MFT, $DATA
-	 *	vol->mft_na
-	 *
-	 * wipe_mft - volume = n mft records, u unused, s total slack space
-	 *	$MFT, $BITMAP
-	 *	vol->mftbmp_na
-	 *
-	 * wipe_directory - volume has d dirs, t total slack space
-	 *	$MFT, $INDEX_ROOT, $INDEX_ALLOC, $BITMAP
-	 *
-	 * wipe_logfile - logfile is <size>
-	 *	$MFT, $DATA
-	 *
-	 * wipe_pagefile - pagefile is <size>
-	 *	$MFT, $DATA
-	 */
-
-	free (buffer);
-#if 0
-	ntfs_inode *inode;
-	ntfs_attr *attr;
-
-	inode = ntfs_inode_open (vol, 6);	/* $Bitmap */
-	if (!inode)
+	/* Get length of pagfile.sys contents. */
+	len = na->data_size;
+	if (!len) {
+		Dprintf("pagefile.sys has zero length, no disk write needed.\n");
 		return 0;
+	}
 
-	attr = ntfs_attr_open (inode, AT_DATA, NULL, 0);
-	if (!attr)
-		return 0;
+	memset(buf, byte, NTFS_BUF_SIZE2);
 
-	ntfs_attr_pread
+	/* Set the $DATA attribute. */
+	pos = 0;
+	while ((count = len - pos) > 0) {
+		if (count > NTFS_BUF_SIZE2)
+			count = NTFS_BUF_SIZE2;
 
-	ntfs_attr_close (attr);
-	ntfs_inode_close (inode);
-#endif
+		if ((count = ntfs_attr_pwrite(na, pos, count, buf)) <= 0) {
+			Dprintf("Failed to set the pagefile.sys attribute value.");
+			if (count != -1)
+				errno = EIO;
+			goto io_error_exit;
+		}
 
-	return 1;
+		pos += count;
+	}
+
+	ntfs_attr_close(na);
+	ntfs_inode_close(ni);
+	Qprintf ("wipe_pagefile 0x%02x, %lld bytes\n", byte, pos);
+	return pos;
+
+io_error_exit:
+	eo = errno;
+	ntfs_attr_close(na);
+	errno = eo;
+error_exit:
+	eo = errno;
+	ntfs_inode_close(ni);
+	errno = eo;
+	return -1;
 }
 
 /**
@@ -933,7 +1230,7 @@ bmpdone:
  *
  * Return:  none
  */
-void print_summary (void)
+static void print_summary (void)
 {
 	int i;
 
@@ -1016,21 +1313,6 @@ int main (int argc, char *argv[])
 		sleep (5);
 	}
 
-	if (0)
-	{
-		int i = 0;
-		runlist_element *rl = vol->mft_na->rl;
-		printf ("________________________________________________________________________________\n\n");
-		for (; rl->length > 0; rl++, i++) {
-			printf ("%4d %lld,%lld,%lld\n", i, (long long)rl->vcn,
-					(long long)rl->lcn,
-					(long long)rl->length);
-		}
-		printf ("%4d %lld,%lld,%lld\n", i, (long long)rl->vcn,
-				(long long)rl->lcn, (long long)rl->length);
-		return 0;
-	}
-
 	printf ("\n");
 	for (i = 0; i < opts.count; i++) {
 		int byte;
@@ -1093,7 +1375,14 @@ int main (int argc, char *argv[])
 
 		printf ("%lld bytes were wiped\n", (long long)total);
 	}
-
+	/*
+	 * We need to reset the logfile so Windows can boot and so journal
+	 * replay does not cause corruption.
+	 */
+	if (act != act_info) {
+		printf ("Resetting logfile.\n");
+		ntfs_logfile_reset(vol);
+	}
 	if (ntfs_volume_set_flags (vol, VOLUME_IS_DIRTY) < 0) {
 		Eprintf ("Couldn't mark volume dirty\n");
 	}
